class Notifier:
    def __init__(self):
        self._start_listeners = set()
        self._update_listeners = set()
        self._completion_listeners = set()
        self._cancel_listeners = set()
        self._pause_listeners = set()
        self._resume_listeners = set()

    def _call_listeners(self, listeners, args, kwargs):
        for listener in listeners:
            listener(*args, **kwargs)

    def add_start_listener(self, listener):
        self._start_listeners.add(listener)

    def remove_start_listener(self, listener):
        self._start_listeners.remove(listener)

    def _call_start_listeners(self, *args, **kwargs):
        self._call_listeners(self._start_listeners, args, kwargs)

    def add_update_listener(self, listener):
        self._update_listeners.add(listener)

    def remove_update_listener(self, listener):
        self._update_listeners.remove(listener)

    def _call_update_listeners(self, *args, **kwargs):
        self._call_listeners(self._update_listeners, args, kwargs)

    def add_completion_listener(self, listener):
        self._completion_listeners.add(listener)

    def remove_completion_listener(self, listener):
        self._completion_listeners.remove(listener)

    def _call_completion_listeners(self, *args, **kwargs):
        self._call_listeners(self._completion_listeners, args, kwargs)

    def add_cancel_listener(self, listener):
        self._cancel_listeners.add(listener)

    def remove_cancel_listener(self, listener):
        self._cancel_listeners.remove(listener)

    def _call_cancel_listeners(self, *args, **kwargs):
        self._call_listeners(self._cancel_listeners, args, kwargs)

    def add_pause_listener(self, listener):
        self._pause_listeners.add(listener)

    def remove_pause_listener(self, listener):
        self._pause_listeners.remove(listener)

    def _call_pause_listeners(self, *args, **kwargs):
        self._call_listeners(self._pause_listeners, args, kwargs)

    def add_resume_listener(self, listener):
        self._resume_listeners.add(listener)

    def remove_resume_listener(self, listener):
        self._resume_listeners.remove(listener)

    def _call_resume_listeners(self, *args, **kwargs):
        self._call_listeners(self._resume_listeners, args, kwargs)



from cellsepi.backend.main_window.data_util import load_directory, ReturnTypePath
from cellsepi.backend.main_window.expert_mode.module import *
from cellsepi.frontend.main_window.gui_directory import DirectoryCard


class ReadLif(Module,ABC):
    _gui_config = ModuleGuiConfig("ReadLif",Categories.INPUTS,"This module handles the read in of .lif files and if available reads in the mask of the images.")
    def __init__(self, module_id: str) -> None:
        super().__init__(module_id)
        self.outputs = {
            "image_paths": OutputPort("image_paths", dict),
            "mask_paths": OutputPort("mask_paths", dict),
        }
        self.user_file_path: FilePath = FilePath(suffix=["lif"])
        self.user_channel_prefix: str = "c"
        self.user_mask_suffix: str = "_seg"

    def run(self):
        working_directory = DirectoryCard().select_directory_parallel(self.user_file_path.path, True, self.user_channel_prefix, self.event_manager)
        self.outputs["image_paths"].data,self.outputs["mask_paths"].data= load_directory(working_directory, self.user_channel_prefix, self.user_mask_suffix, ReturnTypePath.BOTH_PATHS, self.event_manager)


import os
import subprocess
import sys
import platform

def build():
    """
    Build cellsepi with flet build platform -v
    """
    #goto cellsepi folder
    project_dir = os.path.dirname(os.path.abspath(__file__))
    os.chdir(project_dir)

    #check which system is used
    system = platform.system().lower()
    if system == "windows":
        target = "windows"
    elif system == "darwin":
        target = "macos"
    elif system == "linux":
        target = "linux"
    else:
        print("Flet build failed: unknown system")
        sys.exit(1)

    #try to build cellsepi
    try:
            subprocess.check_call([sys.executable, "-m", "flet.cli", "build", target, "-v"])
    except subprocess.CalledProcessError as e:
        print(f"Flet build failed: {e}")
        sys.exit(1)



import numpy as np

from src.cellsepi.backend.drawing_window.drawing_util import mask_shifting
import pytest

mask_data = {
        "masks": np.array([
            [0, 1, 2, 3],
            [4, 5, 0, 6],
            [7, 0, 0, 8]
        ]),
        "outlines": np.array([
            [0, 1, 0, 2],
            [3, 0, 4, 5],
            [6, 0, 7, 8]
        ])
    }
mask = mask_data["masks"]
outline = mask_data["outlines"]

def test_drawing_util():

    mask_data_after_delete = {
        "masks": np.array([
            [0, 1, 2, 3],
            [4, 0, 0, 5],
            [6, 0, 0, 7]
        ]),
        "outlines": np.array([
            [0, 1, 0, 2],
            [3, 0, 4, 0],
            [5, 0, 6, 7]
        ])
    }
    cell_id = 5
    cell_mask = (mask == cell_id).copy()
    cell_outline = (outline == cell_id).copy()
    mask[cell_mask] = 0
    outline[cell_outline] = 0
    mask_shifting(mask_data,cell_id)
    assert np.array_equal(mask_data["masks"], mask_data_after_delete["masks"]), "Mask shifting failed for 'masks'"
    assert np.array_equal(mask_data["outlines"],mask_data_after_delete["outlines"]), "Mask shifting failed for 'outlines'"

def test_drawing_util_invalid():
    cell_id = 0
    cell_mask = (mask == cell_id).copy()
    cell_outline = (outline == cell_id).copy()
    mask[cell_mask] = 0
    outline[cell_outline] = 0
    with pytest.raises(ValueError):
        mask_shifting(mask_data,cell_id)




import flet as ft
fluorescence_button= ft.ElevatedButton(text= "Readout",
                                       icon=ft.Icons.FILE_DOWNLOAD,
                                       tooltip="Readout fluorescence values",
                                       disabled=False,
                                       visible=False)

def error_banner(gui, message):
    gui.page.open(ft.SnackBar(
        ft.Text(message)))
    gui.page.update()


from abc import ABC, abstractmethod
from enum import Enum
from typing import List, Type, TypeVar, Generic

class Event(ABC):
    """
    Abstract base class for all event types.
    Subclass this to define specific event types.
    """
    pass

class ProgressEvent(Event):
    def __init__(self,percent: int,process: str):
        self.percent = percent
        self.process = process

    def __str__(self):
        return f"ProgressEvent: {self.percent}% - {self.process}"

class OnPipelineChangeEvent(Event):
    def __init__(self,change_type: str):
        self.change_type = change_type

class ModuleExecutedEvent(Event):
    def __init__(self,module_id: str):
        self.module_id = module_id

class ModuleStartedEvent(Event):
    def __init__(self,module_id: str):
        self.module_id = module_id

class ErrorEvent(Event):
    def __init__(self,error_name: str, error_msg: str):
        self.error_name = error_name
        self.error_msg = error_msg

    def __str__(self):
        return f"Error_name: {self.error_name} Error_msg: {self.error_msg}"

class DragAndDropEvent(Event):
    def __init__(self,drag:bool):
        self.drag = drag #False if it is no longer dragging or not valid dragging

class PipelinePauseEvent(Event):
    def __init__(self,module_id: str,resume: bool= False):
        self.module_id = module_id
        self.resume = resume

class PipelineCancelEvent(Event):
    def __init__(self,module_id: str):
        self.module_id = module_id

class PipelineErrorEvent(Event):
    def __init__(self,error_name: str,error_msg:str):
        self.error_name = error_name
        self.error_msg = error_msg

class EventListener(ABC):
    """
    Abstract base class for event listeners.
    Subclasses must define the type of event they handle and implement the update logic.
    """
    @abstractmethod
    def get_event_type(self) -> Type[Event]: #pragma: no cover
        pass

    def update(self, event: Event) -> None:
        if not isinstance(event, self.get_event_type()):
            raise TypeError("The given event is not the right event type!")
        self._update(event)

    @abstractmethod
    def _update(self, event: Event) -> None: #pragma: no cover
        pass



import math
import time
from threading import RLock

import flet as ft
from typing import List, Dict

from flet_core import canvas

from cellsepi.frontend.main_window.expert_mode.expert_constants import MODULE_WIDTH, ARROW_PADDING, MODULE_HEIGHT, \
    BUILDER_WIDTH, BUILDER_HEIGHT, \
    ARROW_COLOR, ARROW_LENGTH, ARROW_ANGLE, VALID_COLOR, THROTTLE_UPDATE_LINES
from cellsepi.frontend.main_window.expert_mode.gui_module import ModuleGUI


def calc_angle(x1, y1, x2, y2):
    """
    Calculate the angle between two points in radiant.
    """
    delta_x = x2 - x1
    delta_y = y2 - y1
    return math.atan2(delta_y, delta_x)

def calc_line_points_outside_modules(source_x, source_y, target_x, target_y, padding: float = 0):
    """
    Calculate the points placed outside modules between two points with optional padding for the target side.
    """
    def rect_sides(x, y, target:bool=False):
        w = MODULE_WIDTH / 2 + (padding if target else 0)
        h = MODULE_HEIGHT / 2 + (padding if target else 0)
        return x - w, x + w, y - h, y + h

    def intersect_line_rect(x_1, y_1, x_2, y_2, x_min, x_max, y_min, y_max):
        dx = x_2 - x_1
        dy = y_2 - y_1
        points = []

        if dx != 0:
            t = (x_min - x_1) / dx
            y = y_1 + t * dy
            if y_min <= y <= y_max:
                points.append((x_min, y))

            t = (x_max - x_1) / dx
            y = y_1 + t * dy
            if y_min <= y <= y_max:
                points.append((x_max, y))

        if dy != 0:
            t = (y_min - y_1) / dy
            x = x_1 + t * dx
            if x_min <= x <= x_max:
                points.append((x, y_min))

            t = (y_max - y_1) / dy
            x = x_1 + t * dx
            if x_min <= x <= x_max:
                points.append((x, y_max))

        if not points:
            return 0, 0
        points.sort(key=lambda p: (p[0] - x_1) ** 2 + (p[1] - y_1) ** 2)
        return points[0]

    target_x_min, target_x_max, target_y_min, target_y_max = rect_sides(target_x, target_y, True)
    source_x_min, source_x_max, source_y_min, source_y_max = rect_sides(source_x, source_y, False)

    target_point = intersect_line_rect(source_x, source_y, target_x, target_y, target_x_min, target_x_max, target_y_min, target_y_max)
    source_point = intersect_line_rect(target_x, target_y, source_x, source_y, source_x_min, source_x_max, source_y_min, source_y_max)

    return target_point, source_point


def calc_middle_point(start_point_x,start_point_y,arrow_end_x, arrow_end_y):
    """
    Calculate the middle point between two points.
    """
    return (start_point_x+arrow_end_x)/2, (start_point_y + arrow_end_y)/2

class LinesGUI(canvas.Canvas):
    def __init__(self, pipeline_gui):
        super().__init__()
        self.shapes = []
        self.connections: Dict[(str, str),Dict[str]] = {}
        self.pipeline_gui = pipeline_gui
        self.width = BUILDER_WIDTH
        self.height = BUILDER_HEIGHT
        self.expand = True
        self._lock = RLock()
        self._last_update_per_module = {}

    def update_line(self,source_module_gui: ModuleGUI ,target_module_gui: ModuleGUI,ports: List[str]):
        """
        Adds a line between two modules or updates them if it already exists.
        """
        key = (source_module_gui.module_id, target_module_gui.module_id)

        if key in self.connections:
            del self.connections[key]

        source_x = source_module_gui.left + (MODULE_WIDTH / 2)
        source_y = source_module_gui.top + (MODULE_HEIGHT / 2)
        target_x = target_module_gui.left + (MODULE_WIDTH / 2)
        target_y = target_module_gui.top+ (MODULE_HEIGHT / 2)
        edge = canvas.Line(
            x1=source_x, y1=source_y,
            x2=target_x, y2=target_y,
            paint=ft.Paint(stroke_width=3, color=ARROW_COLOR)
        )
        edge_angle = calc_angle(source_x,source_y,target_x,target_y)

        (target_x_outside,target_y_outside),(source_x_outside,source_y_outside) = calc_line_points_outside_modules(
            source_x, source_y, target_x, target_y, padding=ARROW_PADDING)

        arrow_line_x1 = target_x_outside - ARROW_LENGTH * math.cos(edge_angle - ARROW_ANGLE)
        arrow_line_y1 = target_y_outside - ARROW_LENGTH * math.sin(edge_angle - ARROW_ANGLE)

        arrow_line_x2 = target_x_outside - ARROW_LENGTH * math.cos(edge_angle + ARROW_ANGLE)
        arrow_line_y2 = target_y_outside - ARROW_LENGTH * math.sin(edge_angle + ARROW_ANGLE)

        arrow_end_x = (arrow_line_x1 + arrow_line_x2)/2 #End is the Flat side of the Arrow
        arrow_end_y = (arrow_line_y1 + arrow_line_y2)/2
        port_x,port_y = calc_middle_point(source_x_outside, source_y_outside, arrow_end_x, arrow_end_y)

        arrow = canvas.Path(
                [
                    canvas.Path.MoveTo(target_x_outside,target_y_outside),
                    canvas.Path.LineTo(arrow_line_x1,arrow_line_y1),
                    canvas.Path.LineTo(arrow_line_x2, arrow_line_y2),
                    canvas.Path.Close()
                ],
                paint=ft.Paint(
                    style=ft.PaintingStyle.FILL,color=ARROW_COLOR
                ),
            )

        port_str = ", ".join(ports)
        port_txt = canvas.Text(
            port_x,port_y,
            str(port_str),max_width=220, style=ft.TextStyle(size=15,weight=ft.FontWeight.BOLD,bgcolor=ft.Colors.WHITE38),
            alignment=ft.alignment.center,visible=self.pipeline_gui.show_ports
        )

        def dummy():
            pass
        disabled = False
        opacity = 1
        if (source_module_gui.module_id in self.pipeline_gui.pipeline.run_order or target_module_gui.module_id in self.pipeline_gui.pipeline.run_order or source_module_gui.module_id == self.pipeline_gui.pipeline.executing or target_module_gui.module_id == self.pipeline_gui.pipeline.executing) and self.pipeline_gui.pipeline.running:
            disabled = True
            opacity = 0.4
        delete_button = ft.GestureDetector(top=port_y-20,left=port_x-20,on_hover=lambda e:dummy(),content=ft.IconButton(
            icon=ft.Icons.CLOSE,tooltip="Delete Connection",hover_color=VALID_COLOR,icon_color=ft.Colors.WHITE,bgcolor=ft.Colors.RED_ACCENT,opacity=opacity,on_click=lambda e,source=source_module_gui,target=target_module_gui:self.pipeline_gui.remove_connection(source,target)
            ),visible=self.pipeline_gui.show_delete_button,disabled=disabled)

        self.connections[key] = {
            "edge": edge,
            "arrow": arrow,
            "port_txt": port_txt,
            "delete_button": delete_button,
        }

    def update_delete_button(self,source_module_gui: ModuleGUI, target_module_gui: ModuleGUI,set_all: bool = False):
        """
        Checks and updates the delete button for the connection between the source module and the target module.
        """
        disabled = False
        opacity = 1
        if ((source_module_gui.module_id in self.pipeline_gui.pipeline.run_order or target_module_gui.module_id in self.pipeline_gui.pipeline.run_order or source_module_gui.module_id == self.pipeline_gui.pipeline.executing or target_module_gui.module_id == self.pipeline_gui.pipeline.executing) and self.pipeline_gui.pipeline.running) or set_all:
            disabled = True
            opacity = 0.4
        key = (source_module_gui.module_id,target_module_gui.module_id)
        if key in self.connections:
            delete_button = self.connections[key]["delete_button"]
            delete_button.content.opacity = opacity
            delete_button.content.disabled = disabled
            delete_button.content.update()

    def update_delete_buttons(self,module_gui: ModuleGUI,set_all: bool = False):
        """
        Updates all delete buttons that are connected to the given module.
        """
        for pipe in self.pipeline_gui.pipeline.pipes_in[module_gui.module.module_id]:
            self.update_delete_button(self.pipeline_gui.modules[pipe.source_module.module_id], module_gui,set_all)
        for pipe in self.pipeline_gui.pipeline.pipes_out[module_gui.module.module_id]:
            self.update_delete_button(module_gui, self.pipeline_gui.modules[pipe.target_module.module_id],set_all)

    def remove_line(self, source_module_gui: ModuleGUI, target_module_gui: ModuleGUI):
        """
        Removes a line between two modules.
        """
        key = (source_module_gui.module_id, target_module_gui.module_id)
        conn = self.connections.pop(key, None)
        if conn is not None:
            with self._lock:
                self.shapes.remove(conn["edge"])
                self.shapes.remove(conn["arrow"])
                self.shapes.remove(conn["port_txt"])
                self.pipeline_gui.delete_stack.controls.remove(conn["delete_button"])
                self.update()
                self.pipeline_gui.delete_stack.update()

    def _update_lines(self,module_gui: ModuleGUI):
        """
        Updates all lines that are connected to the given module.
        """
        for pipe in self.pipeline_gui.pipeline.pipes_in[module_gui.module_id]:
            self.update_line(self.pipeline_gui.modules[pipe.source_module.module_id],module_gui,pipe.ports)
        for pipe in self.pipeline_gui.pipeline.pipes_out[module_gui.module_id]:
            self.update_line(module_gui,self.pipeline_gui.modules[pipe.target_module.module_id],pipe.ports)

    def update_lines(self,module_gui: ModuleGUI):
        """
        Updates all lines connected to the given module,
        but only if enough time has passed since the last update
        (for throttling purposes to improve performance during drag/move).
        """
        now = time.time()
        last = self._last_update_per_module.get(module_gui.module_id, 0)
        if now - last > THROTTLE_UPDATE_LINES:
            self._last_update_per_module[module_gui.module_id] = now
            self._update_lines(module_gui)
            self.update_gui()

    def update_gui(self):
        """
        Updates the GUI to reflect the current connections.
        Uses a lock to ensure thread-safety while rebuilding the newest state.
        """
        with self._lock:
            self.shapes.clear()
            self.pipeline_gui.delete_stack.controls.clear()
            for key, data in self.connections.copy().items():
                self.shapes.append(data["edge"])
                self.shapes.append(data["arrow"])
                self.shapes.append(data["port_txt"])
                self.pipeline_gui.delete_stack.controls.append(data["delete_button"])
            self.update()
            self.pipeline_gui.delete_stack.update()

    def update_all(self):
        """
        Updates all connections for every module in the pipeline.
        """
        for module in self.pipeline_gui.modules.values():
            self._update_lines(module)
        self.update_gui()


from cellsepi.backend.main_window.images import BatchImageSegmentation
from cellsepi.backend.main_window.notifier import Notifier


class Segmentation(Notifier):
    """
    This class handles connection of GUISegmentation and BatchImageSegmentation.
    """
    def __init__(self, gui_seg, gui):
        super().__init__()
        self.gui = gui
        self.gui_seg = gui_seg
        self.device = "cpu"
        self.batch_image_segmentation = BatchImageSegmentation(self, self.gui, self.device)
        self.batch_image_segmentation.add_start_listener(listener=self.start)
        self.batch_image_segmentation.add_update_listener(listener=self.update)
        self.batch_image_segmentation.add_completion_listener(listener=self.finished)

    # methods, that communicate a change in the state of the segmentation between GUISegmentation and BatchImageSegmentation
    def to_be_cancelled(self):
        self.batch_image_segmentation.cancel_action()
        self.gui.csp.segmentation_running = False

    def to_be_paused(self):
        self.batch_image_segmentation.pause_action()
        self.gui.csp.segmentation_running = False

    def to_be_resumed(self):
        self.batch_image_segmentation.resume_action()
        self.gui.csp.segmentation_running = True

    def is_resuming(self):
        self._call_resume_listeners()

    def finished(self):
        self._call_completion_listeners()

    def update(self, progress, current_image):
        self._call_update_listeners(progress, current_image)

    def start(self):
        current_percentage = round(self.batch_image_segmentation.num_seg_images / len(self.gui.csp.image_paths) * 100)
        self._call_update_listeners(str(current_percentage) + " %", None)

    def run(self):
        """
        This method starts the segmentation process and manages the different interactions.
        """
        if not self.gui_seg.segmentation_cancelling and not self.gui_seg.segmentation_pausing and not self.gui_seg.segmentation_resuming:
            self._call_update_listeners("Preparing segmentation", None)

        if self.gui.csp.segmentation_running and not self.gui_seg.segmentation_resuming:
            self._call_completion_listeners()
            return

        self.gui.csp.segmentation_running = True

        self.batch_image_segmentation.run()
        self.gui.csp.segmentation_running = False
        if self.gui_seg.segmentation_cancelling:
            self._call_cancel_listeners()
            return
        elif self.gui_seg.segmentation_pausing:
            self._call_pause_listeners()
            return
        else:
            self._call_completion_listeners()






from collections import defaultdict

import flet as ft
from flet_contrib.color_picker import ColorPicker
from enum import Enum

from cellsepi.frontend.main_window.gui_mask import handle_mask_update
from cellsepi.frontend.main_window.gui_page_overlay import PageOverlay
from cellsepi.frontend.main_window.expert_mode.expert_constants import ModuleType

def hex_to_rgb(hex_color):
    """
    Converts a hex color string to rgb color

    Args:
        hex_color (str)

    Returns:
        rgb_color (tuple)
    """
    hex_color = hex_color.lstrip("#")
    return tuple(int(hex_color[i:i+2], 16) for i in (0, 2, 4))

def rgb_to_hex(rgb_color):
    """
    Converts a rgb color to hex color
    Args:
        rgb_color (tuple)
    Returns:
        hex_color (str)
    """
    return "#{:02x}{:02x}{:02x}".format(*rgb_color)

class ColorTypes(Enum):
    Mask = 1
    Outline = 2

class ColorSelection:
    def __init__(self,gui):
        self.config = gui.csp.config
        self.gui = gui
        color_mask = rgb_to_hex(self.config.get_mask_color())
        color_outline = rgb_to_hex(self.config.get_outline_color())
        self.color_picker = ColorPicker(color=color_mask, width=430)
        self.color_picker.hex.border_color=ft.Colors.BLUE_ACCENT
        self.color_picker.r.border_color=ft.Colors.BLUE_ACCENT
        self.color_picker.g.border_color=ft.Colors.BLUE_ACCENT
        self.color_picker.b.border_color=ft.Colors.BLUE_ACCENT
        self.color_icon_mask = ft.IconButton(icon=ft.Icons.BRIGHTNESS_1_ROUNDED,icon_color=color_mask,disabled=True,mouse_cursor=ft.MouseCursor.CLICK)
        self.color_icon_outline = ft.IconButton(icon=ft.Icons.BRIGHTNESS_1_ROUNDED, icon_color=color_outline,disabled=True,mouse_cursor=ft.MouseCursor.CLICK)
        self.color_type = None
        self.dialog = PageOverlay(self.gui.page,ft.Stack([ft.Row([
            ft.Column([ft.Card(content=ft.Stack([ft.Container(ft.ListTile(height=370,width=430),padding=10),ft.Container(ft.Column(
                [self.color_picker,
                ft.Container(ft.Row([ft.TextButton("Save", on_click=self.change_color)
                 ],alignment=ft.MainAxisAlignment.END),padding=10)
                ]
            ),height=430,width=450,padding=10)]))],horizontal_alignment=ft.CrossAxisAlignment.CENTER,alignment=ft.MainAxisAlignment.CENTER)],alignment=ft.MainAxisAlignment.CENTER)]),
            on_dismiss=self.close_dialog,
        )
    def open_color_picker_mask(self,e):
        self.dialog.open()
        self.color_picker.color = rgb_to_hex(self.config.get_mask_color())
        self.color_type = ColorTypes.Mask
        e.control.page.update()


    def open_color_picker_outline(self, e):
        self.dialog.open()
        self.color_picker.color = rgb_to_hex(self.config.get_outline_color())
        self.color_type = ColorTypes.Outline
        e.control.page.update()

    def change_color(self, e):
        """
        Standard color: Mask outline= green, Filling: red
            if it is reasonable, change the color to the liking
        Attributes:
            self.color_picker (ColorPicker)
        """
        if self.color_type == ColorTypes.Mask:
            self.color_icon_mask.icon_color = self.color_picker.color
            self.config.set_mask_color(hex_to_rgb(self.color_picker.color))
            self.gui.mask.mask_outputs = defaultdict(dict)
            handle_mask_update(self.gui)
            self.gui.queue.put(("color_change", self.config.get_mask_color(), self.config.get_outline_color()))
            ModuleType.REVIEW.value.mask_color = self.config.get_mask_color()
            ModuleType.REVIEW.value.update_class()
        else:
            self.color_icon_outline.icon_color = self.color_picker.color
            self.config.set_outline_color(hex_to_rgb(self.color_picker.color))
            self.gui.mask.mask_outputs = defaultdict(dict)
            handle_mask_update(self.gui)
            self.gui.queue.put(("color_change", self.config.get_mask_color(), self.config.get_outline_color()))
            ModuleType.REVIEW.value.outline_color = self.config.get_outline_color()
            ModuleType.REVIEW.value.update_class()
        self.dialog.close()

    def close_dialog(self, e):
        self.dialog.close()

class ColorOpacity:
    def __init__(self,gui):
        self.gui = gui
        self.slider = ft.Slider(
            min=10, max=128, value=128, width=142,
            on_change=lambda _: self.opacity_change()
        )
        self.text = ft.Container(
            content=ft.Text("Mask Opacity"),
            alignment=ft.alignment.center
        )

    def opacity_change(self):
        self.gui.csp.color_opacity = self.slider.value
        self.gui.mask.mask_outputs = defaultdict(dict)
        handle_mask_update(self.gui)
        ModuleType.REVIEW.value.mask_opacity = self.slider.value
        ModuleType.REVIEW.value.update_class()


import math
import platform
from enum import Enum
import flet as ft

from cellsepi.backend.main_window.expert_mode.modules.image_readout import ImageReadoutModule
from cellsepi.backend.main_window.expert_mode.modules.image_segmentation import ImageSegmentationModule
from cellsepi.backend.main_window.expert_mode.modules.project_3d_to_2d import Project3dTo2d
from cellsepi.backend.main_window.expert_mode.modules.read_lif import ReadLif
from cellsepi.backend.main_window.expert_mode.modules.read_tif import ReadTif
from cellsepi.backend.main_window.expert_mode.modules.review import Review
from cellsepi.backend.main_window.expert_mode.modules.spot_detection import SpotDetectionModule

#Constants used in the PipelineBuildingTool(ExpertMode)
BUILDER_WIDTH = 1000
BUILDER_HEIGHT = 400
MODULE_WIDTH = 235
MODULE_HEIGHT = 80
SPACING_X = 10
SPACING_Y = 20
SHOWROOM_PADDING_X = 20
SHOWROOM_MODULE_COUNT = 4
ARROW_LENGTH = 23
ARROW_ANGLE = math.radians(40)
ARROW_PADDING = -1
ARROW_COLOR = ft.Colors.CYAN_900
INVALID_COLOR = ft.Colors.BLACK54
VALID_COLOR = ft.Colors.WHITE30
MENU_COLOR = ft.Colors.BLACK54
DISABLED_BUTTONS_COLOR = ft.Colors.BLACK12
USER_OPTIONS_LIMIT = 9
ZOOM_VALUE = 0.20
BOTTOM_SPACING = 20
MAIN_ACTIVE_COLOR = ft.Colors.WHITE60
THROTTLE_UPDATE_LINES = 0.036 #~30FPS
DEBUG = False

class ModuleType(Enum):
    """
    Enum for all Modules.
    Register new Modules here!
    """
    IMAGE_READOUT = ImageReadoutModule
    IMAGE_SEGMENTATION = ImageSegmentationModule
    SPOT_DETECTION = SpotDetectionModule
    READ_LIF = ReadLif
    READ_TIF = ReadTif
    PROJECTION_3D_TO_2D = Project3dTo2d
    REVIEW = Review





from os import path

from cellsepi.backend.main_window.data_util import load_directory, ReturnTypePath
from cellsepi.backend.main_window.expert_mode.listener import ProgressEvent
from cellsepi.backend.main_window.expert_mode.module import *
from cellsepi.backend.main_window.images import BatchImageSegmentation, BatchImageReadout
from cellsepi.frontend.main_window.gui_directory import DirectoryCard


class ImageReadoutModule(Module, ABC):
    _gui_config = ModuleGuiConfig("ImageReadout",Categories.OUTPUTS,"This module handles the readout of the segmented images and saves them in an .xlsx file.")
    def __init__(self, module_id: str) -> None:
        super().__init__(module_id)
        self.inputs = {
            "image_paths": InputPort("image_paths", dict),
            "mask_paths": InputPort("mask_paths", dict),
        }
        self.user_directory_path: DirectoryPath = DirectoryPath()
        self.user_segmentation_channel: str = "2"
        self.user_channel_prefix: str = "c"

    def run(self):
        BatchImageReadout(self.inputs["image_paths"].data, self.inputs["mask_paths"].data,self.user_segmentation_channel,self.user_channel_prefix,self.user_directory_path.path,True).run(self.event_manager)


import asyncio
import textwrap
from typing import List, Any, Dict

from flet_core.cupertino_colors import WHITE

from cellsepi.backend.main_window.expert_mode.listener import DragAndDropEvent, OnPipelineChangeEvent
from cellsepi.backend.main_window.expert_mode.module import FilePath, DirectoryPath
from cellsepi.frontend.main_window.expert_mode.expert_constants import *
from cellsepi.frontend.main_window.gui_directory import format_directory_path
from cellsepi.frontend.main_window.gui_page_overlay import PageOverlay


class ModuleGUI(ft.GestureDetector):
    """
    Manages the GUI parts of the module.
    """
    def __init__(self, pipeline_gui,module_type: ModuleType,x: float = None,y: float = None,show_mode:bool=False,visible=True,index:int=None,id_number:int=None,module_dict:dict=None):
        super().__init__()
        self.pipeline_gui = pipeline_gui
        self.detection: bool = True
        self.module_type = module_type
        self.mouse_cursor = ft.MouseCursor.MOVE
        self.show_mode = show_mode
        self.drag_interval = 5
        self.visible = visible
        self.on_pan_start = self.start_drag
        self.on_pan_update = self.drag
        self.on_pan_end = self.drop
        self.show_offset_y = y
        self.left = BUILDER_WIDTH/2 if x is None else x
        self.top = BUILDER_HEIGHT/2 if y is None else y
        self.old_left = None
        self.old_top = None
        self.port_selection = False
        self.module = self.pipeline_gui.pipeline.add_module(module_type.value) if id_number is None else self.pipeline_gui.pipeline.add_module_with_id(module_type.value,module_type.value.gui_config().name + str(id_number))
        if module_dict is not None:
            self.update_user_attr(module_dict)
        if self.module.settings is None and hasattr(self.module, "_settings"):
            self.module._settings = self.generate_options_overlay()
        if show_mode:
            if index is None:
                self.pipeline_gui.show_room_modules.append(self)
            else:
                self.pipeline_gui.show_room_modules.insert(index,self)
        else:
            self.pipeline_gui.modules[self.module.module_id] = self
            self.pipeline_gui.pipeline.event_manager.notify(OnPipelineChangeEvent(f"Added {self.module_id} to pipeline_gui.modules."))
        self.color = self.module.gui_config().category.value
        self.valid = False
        self.wrapped_description = "\n".join(textwrap.wrap(self.module.gui_config().description, width=40))
        self.click_container = ft.Container(on_click=lambda e: self.add_connection(),tooltip=self.wrapped_description if self.show_mode else None, height=MODULE_HEIGHT, width=MODULE_WIDTH,
                                            visible=False if not show_mode else True,bgcolor=INVALID_COLOR if not show_mode else ft.Colors.TRANSPARENT,disabled=True if not show_mode else False,border_radius=ft.border_radius.all(10))
        self.click_gesture = ft.GestureDetector(visible=False if not show_mode else True,disabled=True if not show_mode else False,content=self.click_container,on_enter=lambda e: self.on_enter_click_module(),on_exit=lambda e: self.on_exit_click_module())
        self.connect_button = ft.IconButton(icon=ft.Icons.SHARE, icon_color=ft.Colors.WHITE60,
                                            style=ft.ButtonStyle(
                                                          shape=ft.RoundedRectangleBorder(radius=12),
                                                      ), on_click=lambda e: self.connect_clicked(),
                                            tooltip="Add connection", hover_color=ft.Colors.WHITE12, visible=self.module.outputs!={})

        self.options_button = ft.IconButton(icon=ft.Icons.TUNE, icon_color=ft.Colors.WHITE60,
                                            style=ft.ButtonStyle(
                                                          shape=ft.RoundedRectangleBorder(radius=12),
                                                      ), on_click=lambda e: self.open_options(e),
                                            tooltip="Options", hover_color=ft.Colors.WHITE12, visible=True if self.module.settings is not None else False,)
        self.copy_button = ft.IconButton(icon=ft.Icons.CONTENT_COPY, icon_color=ft.Colors.WHITE60,
                                            style=ft.ButtonStyle(
                                                          shape=ft.RoundedRectangleBorder(radius=12),
                                                      ), on_click=lambda e: self.copy_module(),
                                            tooltip="Copy module", hover_color=ft.Colors.WHITE12,)
        self.paused_button = ft.Stack([ft.Container(bgcolor=ft.Colors.BLACK26, width=30, height=30, top=5, right=5, border_radius=ft.border_radius.all(45)), ft.IconButton(icon=ft.Icons.PLAY_ARROW, icon_color=ft.Colors.WHITE, disabled=True,
                                                                                                                                                                           style=ft.ButtonStyle(
                                              shape=ft.RoundedRectangleBorder(radius=12),
                                          ), on_click=lambda e: self.copy_module(),
                                                                                                                                                                           tooltip="Currently paused to continue\npress the resume button", hover_color=ft.Colors.WHITE12)], top=1,
                                      left=MODULE_WIDTH - 42, visible=False, width=40, height=40)
        self.executing_button = ft.Stack([ft.Container(bgcolor=ft.Colors.BLACK26, width=30, height=30, top=5, right=5, border_radius=ft.border_radius.all(45)), ft.IconButton(icon=ft.Icons.PAUSE_ROUNDED, icon_color=ft.Colors.WHITE, disabled=True,
                                                                                                                                                                              style=ft.ButtonStyle(
                                              shape=ft.RoundedRectangleBorder(radius=12),
                                          ), on_click=lambda e: self.copy_module(),
                                                                                                                                                                              tooltip="Currently executed", hover_color=ft.Colors.WHITE12)], top=1,
                                         left=MODULE_WIDTH - 42, visible=False, width=40, height=40)
        self.waiting_button = ft.Stack([ft.Container(bgcolor=ft.Colors.BLACK26,width=30,height=30,top=5,right=5,border_radius=ft.border_radius.all(45)),ft.IconButton(icon=ft.Icons.HOURGLASS_EMPTY_ROUNDED, icon_color=ft.Colors.WHITE,disabled=True,
                                          style=ft.ButtonStyle(
                                              shape=ft.RoundedRectangleBorder(radius=12),
                                          ), on_click=lambda e: self.copy_module(),
                                          tooltip="Waiting for execution", hover_color=ft.Colors.WHITE12)],top=1,
                                          left=MODULE_WIDTH - 42,visible=False,width=40,height=40)
        self.show_ports = False
        self.ports_in_out_button = ft.IconButton(icon=ft.Icons.SYNC_ALT_ROUNDED, icon_color=ft.Colors.WHITE60,
                                                 style=ft.ButtonStyle(
                                              shape=ft.RoundedRectangleBorder(radius=12),
                                          ), on_click=lambda e: self.ports_in_out_clicked(),
                                                 tooltip="View ports", hover_color=ft.Colors.WHITE12,)

        self.tools = ft.Container(ft.Row(
                                            [
                                                self.connect_button,self.options_button,self.ports_in_out_button,self.copy_button,
                                            ],tight=True,spacing=7
                                        ),bgcolor=ft.Colors.BLACK12,expand=True,width=MODULE_WIDTH
                                        )

        self.delete_button = ft.IconButton(ft.Icons.CLOSE,visible=True if not show_mode else False,icon_color=ft.Colors.WHITE,hover_color=ft.Colors.WHITE12,tooltip="Delete Module",on_click=lambda e:self.remove_module())
        self.port_chips = self.get_ports_row()
        self.connection_ports = ft.Container(
            self.port_chips,visible=False
        )

        control_list_ports = []
        self.in_ports_Icons = {}
        self.in_ports_Icons_occupied = {}
        for port in self.module.inputs.values():
            if not port.opt:
                self.in_ports_Icons[port.name] = ft.Stack([ft.Container(bgcolor=WHITE,width=10,height=20,bottom=10,right=15,border_radius=ft.border_radius.all(45)),ft.Container(ft.IconButton(ft.Icons.WARNING_ROUNDED,icon_size=35,disabled=True,hover_color=ft.Colors.TRANSPARENT,icon_color=ft.Colors.RED,tooltip=f"Port '{port.name}' is mandatory and has no incoming pipe!"),bottom=-3,right=-5)],alignment=ft.alignment.center,visible=not self.pipeline_gui.pipeline.check_ports_occupied(self.module_id,[port.name]),width=40,height=40)
                self.in_ports_Icons_occupied[port.name] = ft.Stack([ft.Container(bgcolor=ft.Colors.GREEN,width=30,height=30,border_radius=ft.border_radius.all(45)),ft.IconButton(ft.Icons.CHECK,disabled=True,hover_color=ft.Colors.TRANSPARENT,icon_color=WHITE,tooltip=f"Port '{port.name}' is mandatory and is satisfied.")],alignment=ft.alignment.center,visible= self.pipeline_gui.pipeline.check_ports_occupied(self.module_id,[port.name]),width=40,height=40)
            else:
                self.in_ports_Icons[port.name] = ft.Stack([ft.Container(bgcolor=WHITE,width=10,height=20,bottom=10,right=15,border_radius=ft.border_radius.all(45)),ft.Container(ft.IconButton(ft.Icons.WARNING_ROUNDED,icon_size=35,disabled=True,hover_color=ft.Colors.TRANSPARENT,icon_color=ft.Colors.RED,tooltip=f"Port '{port.name}' is optional and has no incoming pipe!"),bottom=-3,right=-5)],alignment=ft.alignment.center,visible=not self.pipeline_gui.pipeline.check_ports_occupied(self.module_id,[port.name]),opacity=0.2,width=40,height=40)
                self.in_ports_Icons_occupied[port.name] = ft.Stack([ft.Container(bgcolor=ft.Colors.GREEN,width=30,height=30,border_radius=ft.border_radius.all(45)),ft.IconButton(ft.Icons.CHECK,disabled=True,hover_color=ft.Colors.TRANSPARENT,icon_color=WHITE,tooltip=f"Port '{port.name}' is optional and is satisfied.")],alignment=ft.alignment.center,opacity=0.2,visible= self.pipeline_gui.pipeline.check_ports_occupied(self.module_id,[port.name]),width=40,height=40)

        in_ports = ft.Column([ft.Row([ft.Text(port.name,width=MODULE_WIDTH/2,weight=ft.FontWeight.BOLD,color=ft.Colors.WHITE),self.in_ports_Icons[port.name],self.in_ports_Icons_occupied[port.name]]) for port in self.module.inputs.values()],spacing=0)
        out_ports = ft.Column([ft.Row([ft.Text(port.name,width=MODULE_WIDTH/2,weight=ft.FontWeight.BOLD,color=ft.Colors.WHITE)]) for port in self.module.outputs.values()])
        input_text=ft.Text("Inputs:",size=20,weight=ft.FontWeight.BOLD,color=ft.Colors.WHITE)
        output_text=ft.Text("Outputs:",size=20,weight=ft.FontWeight.BOLD,color=ft.Colors.WHITE)
        if self.module.inputs != {}:
            control_list_ports.append(input_text)
            control_list_ports.append(in_ports)
        if self.module.outputs != {}:
            control_list_ports.append(output_text)
            control_list_ports.append(out_ports)

        self.error_icon = ft.IconButton(ft.Icons.REPORT, icon_size=35, disabled=True,
                                                         hover_color=ft.Colors.TRANSPARENT, icon_color=ft.Colors.RED,
                                                         tooltip="An error occurred while executing!",
                                                         highlight_color=ft.Colors.TRANSPARENT,
                                                         padding=ft.padding.all(2))
        self.error_stack = ft.Stack([ft.Container(bgcolor=WHITE, width=10, height=20, bottom=16, right=23,
                                                        border_radius=ft.border_radius.all(45)),
                                           self.error_icon],
                                          visible=False, width=48, height=48, top=1,
                                          left=MODULE_WIDTH - 75)

        self.warning_satisfied = ft.Stack([ft.Container(bgcolor=WHITE,width=10,height=20,bottom=16,right=23,border_radius=ft.border_radius.all(45)),ft.IconButton(ft.Icons.WARNING_ROUNDED,icon_size=35,disabled=False,hover_color=ft.Colors.TRANSPARENT,icon_color=ft.Colors.RED,tooltip=f"Not all mandatory inputs are satisfied!",on_click=lambda e:self.ports_in_out_clicked(),highlight_color=ft.Colors.TRANSPARENT,padding=ft.padding.all(2))],visible=not self.pipeline_gui.pipeline.check_module_satisfied(self.module_id) and not show_mode,width=48,height=48,top=1,left=MODULE_WIDTH-75)
        self.name_text = ft.Text(value=self.module.gui_config().name if not DEBUG else self.module_id,
                                weight=ft.FontWeight.BOLD,
                                width=MODULE_WIDTH-80,
                                height=20,color=ft.Colors.BLACK,
                                overflow=ft.TextOverflow.ELLIPSIS)
        self.module_container = ft.Container(
                    content=ft.Column(
                                [
                                        ft.Container(ft.Row(
                                                [
                                                    self.name_text,
                                                ],height=20
                                              ),padding=ft.padding.only(left=5, top=5)),
                                            self.tools,
                                       ],
                                    tight=True)
                    ,bgcolor=self.color,width=MODULE_WIDTH
                    ,height=MODULE_HEIGHT,
                    border=ft.border.all(4, ft.Colors.RED if not self.pipeline_gui.pipeline.check_module_satisfied(self.module_id) and not show_mode else ft.Colors.BLACK12),
                    border_radius=ft.border_radius.all(10),
                )
        self.ports_column = ft.Column(controls=control_list_ports, scroll=ft.ScrollMode.HIDDEN)
        self.ports_container = ft.Container(
            content=self.ports_column, bgcolor=self.color,
            width=MODULE_WIDTH,
            border_radius=ft.border_radius.all(10), padding=10, top=self.module_container.height-15,
            border=ft.border.all(8, ft.Colors.BLACK12), height=0,opacity=0,
            animate_opacity=ft.Animation(duration=300, curve=ft.AnimationCurve.LINEAR_TO_EASE_OUT),
            animate=ft.Animation(duration=300, curve=ft.AnimationCurve.LINEAR_TO_EASE_OUT),
        )
        self.content = ft.Stack([
            self.ports_container,
            ft.Column([ft.Stack([
                self.module_container,
                ft.Container(content=self.delete_button, margin=ft.margin.only(top=-7, left=7),
                             alignment=ft.alignment.top_right, width=MODULE_WIDTH, ),
                self.executing_button,
                self.waiting_button,
                self.paused_button,
                self.warning_satisfied,
                self.error_stack,
                self.click_gesture,
        ]
        ),
        self.connection_ports

        ],tight=True
        )],height=self.module_container.height,
        )
        self.page_overlay = PageOverlay(self.pipeline_gui.page,self.module.settings,self.close_options)
        self._ports_lock = asyncio.Lock()

    def disable_tools(self):
        """
        Disable the tools of the modules.
        """
        self.warning_satisfied.disabled = True
        self.warning_satisfied.update()
        if self.port_selection:
            self.connect_clicked()
        self.connect_button.disabled=True
        self.connect_button.icon_color = DISABLED_BUTTONS_COLOR
        self.options_button.disabled=True
        self.options_button.icon_color = DISABLED_BUTTONS_COLOR
        self.copy_button.disabled = True
        self.copy_button.icon_color = DISABLED_BUTTONS_COLOR
        self.ports_in_out_clicked(disable=True)
        self.connect_button.update()
        self.options_button.update()
        self.copy_button.update()

    def enable_pause(self):
        """
        Enable the options button while be in a pausing state.
        """
        self.options_button.disabled = False
        self.options_button.icon_color = ft.Colors.WHITE60
        self.options_button.update()

    def disable_pause(self):
        """
        Disable the options button after a pausing state.
        """
        self.options_button.disabled = True
        self.options_button.icon_color = DISABLED_BUTTONS_COLOR
        self.options_button.update()

    def enable_tools(self):
        """
        Enable the tools of the module.
        """
        self.warning_satisfied.disabled = False
        self.warning_satisfied.update()
        self.connect_button.disabled = False
        self.connect_button.icon_color = ft.Colors.WHITE60
        self.options_button.disabled = False
        self.options_button.icon_color = ft.Colors.WHITE60
        self.copy_button.disabled = False
        self.copy_button.icon_color = ft.Colors.WHITE60
        self.ports_in_out_clicked(disable=False)
        self.connect_button.update()
        self.options_button.update()
        self.copy_button.update()

    def on_enter_click_module(self):
        """
        Handles if the mouse enters hovering over the module to connect.
        """
        if self.valid:
            self.click_container.bgcolor = VALID_COLOR
            self.click_container.update()

    def on_exit_click_module(self):
        """
        Handles if the mouse exits hovering over the module to connect.
        """
        if self.valid:
            self.click_container.bgcolor = ft.Colors.TRANSPARENT
            self.click_container.update()

    def update_port_icons(self):
        """
        Updates all ports_Icons of the show ports tab.
        """
        for port in self.module.inputs.keys():
            if self.pipeline_gui.pipeline.check_ports_occupied(self.module_id, [port]):
                self.in_ports_Icons[port].visible = False
                self.in_ports_Icons_occupied[port].visible = True
            else:
                self.in_ports_Icons[port].visible = True
                self.in_ports_Icons_occupied[port].visible = False

            self.in_ports_Icons[port].update()
            self.in_ports_Icons_occupied[port].update()

            self.check_warning()

    def check_warning(self):
        """
        Checks if the module should have a warning.
        """
        self.module_container.border = ft.border.all(4,
                                                     ft.Colors.RED if not self.pipeline_gui.pipeline.check_module_satisfied(
                                                         self.module_id) or self.error_stack.visible else ft.Colors.BLACK12)
        self.module_container.update()
        if not self.error_stack.visible:
            self.warning_satisfied.visible = not self.pipeline_gui.pipeline.check_module_satisfied(self.module_id)
            self.warning_satisfied.update()

    def connect_clicked(self,update:bool=True):
        """
        Handles the event when the connection button gets pressed.
        """
        self.pipeline_gui.toggle_all_module_detection(self.module_id)
        if not self.port_selection:
            if self.show_ports:
                self.ports_in_out_clicked(False)
            if update:
                self.pipeline_gui.set_in_background(self,True)
            self.connect_button.icon_color = ft.Colors.BLACK38
            self.delete_button.visible = False
            self.content.height = 40+self.module_container.height
            self.connection_ports.visible = True
            self.connection_ports.update()
            self.port_selection = True
        else:
            self.valid = False
            self.connection_ports.visible = False
            self.connection_ports.update()
            self.port_chips = self.get_ports_row()
            self.connection_ports.content = self.port_chips
            self.connect_button.icon_color = ft.Colors.WHITE60
            self.delete_button.visible = True
            self.content.height = self.module_container.height
            self.port_selection = False
            self.pipeline_gui.source_module = ""

        self.delete_button.update()
        self.content.update()
        self.connect_button.update()
        self.pipeline_gui.page.update()

    def ports_in_out_clicked(self, update: bool = True,disable:bool = None):
        self.pipeline_gui.page.run_task(self.async_ports_in_out_clicked, update=update,disable=disable)

    async def async_ports_in_out_clicked(self, update:bool=True,disable:bool=None):
        """
        Handles the event when the show ports button gets pressed.
        """
        async with self._ports_lock:
            if self.module_id not in self.pipeline_gui.modules:
                return
            if not self.show_ports and disable is None:
                if self.port_selection:
                    self.connect_clicked(False)
                    await asyncio.sleep(0.02)

                self.ports_in_out_button.icon_color = ft.Colors.BLACK38
                self.ports_in_out_button.update()
                self.content.height = MODULE_HEIGHT * 2 + self.module_container.height - 15
                if self.module_id in self.pipeline_gui.modules:
                    self.content.update()
                self.ports_container.height = MODULE_HEIGHT * 2
                self.ports_container.opacity = 1
                if self.module_id in self.pipeline_gui.modules:
                    self.ports_container.update()
                await asyncio.sleep(0.14)
                self.ports_column.scroll=ft.ScrollMode.ALWAYS
                if self.module_id in self.pipeline_gui.modules:
                    self.ports_column.update()
                self.show_ports = True
            else:
                if not disable:
                    self.ports_in_out_button.disabled = False
                    self.ports_in_out_button.icon_color = ft.Colors.WHITE60
                    if self.module_id in self.pipeline_gui.modules:
                        self.ports_in_out_button.update()
                else:
                    self.ports_in_out_button.disabled = True
                    self.ports_in_out_button.icon_color = DISABLED_BUTTONS_COLOR
                    if self.module_id in self.pipeline_gui.modules:
                        self.ports_in_out_button.update()

                self.ports_column.scroll=ft.ScrollMode.HIDDEN
                if self.module_id in self.pipeline_gui.modules:
                    self.ports_column.update()
                self.ports_container.height = 0
                self.ports_container.opacity = 0
                if self.module_id in self.pipeline_gui.modules:
                    self.ports_container.update()
                await asyncio.sleep(0.22)
                self.content.height = self.module_container.height
                if self.module_id in self.pipeline_gui.modules:
                    self.content.update()
                self.show_ports = False


    def set_valid(self):
        """
        Sets a module valid to connect.
        """
        self.valid = True
        self.click_container.bgcolor = ft.Colors.TRANSPARENT
        self.module_container.border = ft.border.all(4, ft.Colors.WHITE38)
        self.module_container.update()
        self.click_container.update()

    def set_invalid(self):
        """
        Sets a module to invalid to connect.
        """
        self.valid = False
        self.click_container.bgcolor = INVALID_COLOR
        self.module_container.border = ft.border.all(4, ft.Colors.RED if not self.pipeline_gui.pipeline.check_module_satisfied(self.module_id) or self.error_stack.visible else ft.Colors.BLACK12)
        self.module_container.update()
        self.click_container.update()

    def set_running(self):
        """
        Called when the module is currently running.
        """
        self.executing_button.visible = True
        self.waiting_button.visible = False
        self.executing_button.update()
        self.waiting_button.update()

    def get_ports_row(self):
        """
        Creates the chip row for the different ports of a module.
        """
        ports_chips = ft.Row()

        for port_name in self.module.outputs.keys():
            ports_chips.controls.append(
                ft.Chip(
                    label=ft.Text(port_name),
                    on_select=lambda e,name = port_name: self.select_port(e,name),
                )
            )
        return ports_chips

    def select_port(self,e, port_name):
        """
        Handles the event if a port gets selected for connecting.
        """
        if e.control.selected:
            self.pipeline_gui.transmitting_ports.append(port_name)
            self.port_chips.update()
        else:
            self.pipeline_gui.transmitting_ports.remove(port_name)
            self.port_chips.update()

        self.pipeline_gui.check_for_valid_all_modules()

    def toggle_detection(self):
        """
        Toggles between the module state 'only moveable' and  'normal mode'
        """
        if self.detection:
            self.detection = False
            self.click_container.disabled = False
            self.click_container.visible = True
            self.click_container.update()
            self.click_gesture.disabled = False
            self.click_gesture.visible = True
            self.click_gesture.update()
            self.delete_button.visible = False
            self.delete_button.update()
        else:
            self.detection = True
            self.set_invalid()
            self.click_container.disabled = True
            self.click_container.visible = False
            self.click_container.update()
            self.click_gesture.disabled = True
            self.click_gesture.visible = False
            self.click_gesture.update()
            if (self.module_id not in self.pipeline_gui.pipeline.run_order and not self.pipeline_gui.pipeline.executing == self.module_id) or not self.pipeline_gui.pipeline.running:
                self.delete_button.visible = True
                self.delete_button.update()
            self.connection_ports.visible = False
            self.connect_button.update()


    def add_connection(self):
        """
        Handles the last step of the adding event when the target gets selected.
        """
        if self.pipeline_gui.source_module is not None and self.pipeline_gui.transmitting_ports is not None and not self.detection and self.valid:
            current_pipe = self.pipeline_gui.pipeline.get_pipe(self.pipeline_gui.source_module,self.module_id)
            if current_pipe is None:
                self.pipeline_gui.add_connection(self.pipeline_gui.modules[self.pipeline_gui.source_module],self,self.pipeline_gui.transmitting_ports)
            else:
                self.pipeline_gui.expand_connection(current_pipe,self.pipeline_gui.transmitting_ports)
            self.pipeline_gui.check_for_valid_all_modules()

    def remove_module(self):
        """
        Removes a module and all its connections.
        """
        for pipe in list(self.pipeline_gui.pipeline.pipes_in[self.module_id]):
            self.pipeline_gui.remove_connection(self.pipeline_gui.modules[pipe.source_module.module_id],self)
        for pipe in list(self.pipeline_gui.pipeline.pipes_out[self.module_id]):
            self.pipeline_gui.remove_connection(self,self.pipeline_gui.modules[pipe.target_module.module_id])

        if self.show_mode:
            self.pipeline_gui.show_room_modules.remove(self)
            self.pipeline_gui.pipeline.remove_module(self.module)
            self.pipeline_gui.page_stack.controls.remove(self)
            self.pipeline_gui.update()
        else:
            self.pipeline_gui.remove_module(self.module_id)

    @property
    def module_id(self):
        """
        Returns the module id of the module.
        """
        return self.module.module_id

    def bounce_back(self):
        """Returns card to its original position"""
        self.left = self.old_left
        self.top = self.old_top
        self.pipeline_gui.lines_gui._update_lines(self)
        self.pipeline_gui.lines_gui.update_gui()
        self.update()

    def start_drag(self, e: ft.DragStartEvent):
        """
        Handles the start of the drag event to save old location to make it possible to bounce back.
        """
        self.old_left = self.left
        self.old_top = self.top
        self.pipeline_gui.lines_gui.update_lines(self)
        self.update()

    def drag(self, e: ft.DragUpdateEvent):
        """
        Handles the drag event.
        """
        if self.show_mode:
            self.click_container.tooltip = None
            self.click_container.update()
            overlap_show_room = not (
                    self.left + MODULE_WIDTH  < self.pipeline_gui.show_room_container.left or
                    self.left > self.pipeline_gui.show_room_container.left + self.pipeline_gui.show_room_container.width or
                    self.top + MODULE_HEIGHT < self.pipeline_gui.show_room_container.top or
                    self.top > self.pipeline_gui.show_room_container.top + self.pipeline_gui.show_room_container.height
            ) and self.show_mode
            if not overlap_show_room:
                self.pipeline_gui.pipeline.event_manager.notify(DragAndDropEvent(True))
            else:
                self.pipeline_gui.pipeline.event_manager.notify(DragAndDropEvent(False))

        self.top = max(0, self.top + e.delta_y)
        self.left = max(0, self.left + e.delta_x)
        self.pipeline_gui.lines_gui.update_lines(self)
        self.update()

    def drop(self,e: ft.DragEndEvent):
        """
        Handles the drop event.
        """
        if self.show_mode:
            self.click_container.tooltip = self.wrapped_description
            self.click_container.update()
            self.pipeline_gui.pipeline.event_manager.notify(DragAndDropEvent(False))
        # calc the left and top values in the pipeline_gui
        offset_x, offset_y, scale = self.pipeline_gui.interactive_view.get_transformation_data()
        check_left = self.left if not self.show_mode else (self.left - offset_x) / scale
        check_top = self.top if not self.show_mode else (self.top - offset_y) / scale

        for module in self.pipeline_gui.modules.values():
            if module is self:
                continue



            overlap = not (
                    check_left + MODULE_WIDTH < module.left or
                    check_left > module.left + MODULE_WIDTH or
                    check_top + MODULE_HEIGHT < module.top or
                    check_top > module.top + MODULE_HEIGHT
            )

            if overlap:
                self.bounce_back()
                self.pipeline_gui.lines_gui.update_lines(self)
                e.control.update()
                return

        #no need to calc cords because show_room_container and module with self.show_mode are in the same page_stack
        overlap_show_room = not(
                self.left + MODULE_WIDTH  < self.pipeline_gui.show_room_container.left or
                self.left > self.pipeline_gui.show_room_container.left + self.pipeline_gui.show_room_container.width or
                self.top + MODULE_HEIGHT < self.pipeline_gui.show_room_container.top or
                self.top > self.pipeline_gui.show_room_container.top + self.pipeline_gui.show_room_container.height
        ) and self.show_mode

        if overlap_show_room:
            self.bounce_back()
            e.control.update()
            return
        elif self.show_mode:
            self.show_mode = False
            index = self.pipeline_gui.show_room_modules.index(self)
            self.pipeline_gui.show_room_modules.remove(self)
            show_room_id = self.module.get_id_number()
            self.pipeline_gui.pipeline.get_new_module_id(self.module_id)
            self.pipeline_gui.modules[self.module_id] = self
            if DEBUG:
                self.name_text.value = self.module_id
            self.pipeline_gui.refill_show_room(self,self.visible,index,show_room_id)
            self.pipeline_gui.page_stack.controls.remove(self)
            self.left = check_left
            self.top = check_top
            self.pipeline_gui.controls.append(self)
            self.click_container.disabled = True
            self.click_container.bgcolor = INVALID_COLOR
            self.click_container.visible = False
            self.click_container.tooltip = None
            self.click_gesture.disabled = True
            self.click_gesture.visible = False
            self.delete_button.visible = True
            if self.pipeline_gui.source_module != "":
                self.toggle_detection()
                self.pipeline_gui.check_for_valid_all_modules()
            self.pipeline_gui.page.update()

        e.control.update()
        self.pipeline_gui.lines_gui.update_lines(self)

    def generate_options_overlay(self):
        """
        Generates with the user attributes tagged with the prefix 'user_' a gui overlay.
        """
        user_attributes = self.module.get_user_attributes
        element_height= 60
        spacing = 10
        padding = 20
        height = element_height* len(user_attributes) + spacing * (len(user_attributes)-1)
        limit_reached = len(user_attributes) > USER_OPTIONS_LIMIT
        if len(user_attributes) != 0:
            return ft.Stack([ft.Row([ft.Column([ft.Card(
                content=ft.Column(
                    [ft.Container(ft.ListView(
                        controls=self.create_attribute_list(user_attributes),
                        width=500,
                        spacing=spacing,
                        height=(element_height* (USER_OPTIONS_LIMIT+1) + spacing * ((USER_OPTIONS_LIMIT+1)-1))-30  if limit_reached else height,
                    ),padding=padding)]
                )
            )],
            alignment=ft.MainAxisAlignment.CENTER,)],alignment=ft.MainAxisAlignment.CENTER)],alignment=ft.alignment.center)
        else:
            return None

    def create_attribute_list(self,attributes=None):
        """
        Creates a text field for each user attribute and combines them into a single list.
        Allowed types are:
            int
            float
            string
            boolean
            FilePath
            DirectoryPath
        """
        items = []
        for attribute_name in attributes:
            value = getattr(self.module, attribute_name)
            typ = type(value)
            if typ in (int, float, str):
                ref = ft.Ref[ft.Text]()
                setattr(self.module,"ref_" + attribute_name, ref)
                items.append(ft.TextField(
                            label=attribute_name.removeprefix("user_"),
                            border_color=ft.Colors.BLUE_ACCENT,
                            value=str(value),
                            ref=ref,
                            on_blur=lambda e,attr_name= attribute_name,reference=ref,type_atr = typ: self.on_change(e,attr_name,reference,type_atr),
                            height=60,
                        ))
            elif typ == bool:
                text = ft.Text(attribute_name.removeprefix("user_"),weight=ft.FontWeight.BOLD)
                index = 0 if not value else 1
                setattr(self.module,"on_change_" + attribute_name, lambda: None)
                slider_bool = ft.CupertinoSlidingSegmentedButton(
                    selected_index=index,
                    thumb_color=ft.Colors.BLUE_400,
                    on_change=lambda e, attr_name=attribute_name: self.update_bool(e,attr_name),
                    padding=ft.padding.symmetric(0, 0),
                    controls=[
                        ft.Text("False"),
                        ft.Text("True")
                    ],
                )
                bool_chosing = ft.Container(ft.Row([text,slider_bool],alignment=ft.MainAxisAlignment.SPACE_BETWEEN),padding=ft.Padding(0,10,0,10))
                items.append(bool_chosing)
            elif typ == FilePath:
                text_field = ft.TextField(
                    label=attribute_name.removeprefix("user_"),
                    border_color=ft.Colors.BLUE_ACCENT,
                    value=format_directory_path(value.path,50),
                    height=60,
                    read_only=True,
                    disabled=True
                )
                attr = getattr(self.module, attribute_name)
                file_picker = ft.FilePicker(on_result=lambda a,attr_name= attribute_name,text=text_field: self.on_select_file(a,attr_name,text))
                self.pipeline_gui.page.overlay.extend([file_picker])
                items.append(ft.Stack([text_field,ft.Container(
                                content=ft.IconButton(
                                    icon=ft.Icons.UPLOAD_FILE,
                                    tooltip="Pick file",
                                    on_click=lambda e: file_picker.pick_files(allow_multiple=False,file_type=ft.FilePickerFileType.CUSTOM if attr.suffix is not None else ft.FilePickerFileType.ANY,allowed_extensions=attr.suffix if attr.suffix is not None else [],),
                                ),
                                alignment=ft.alignment.top_right
                            ,right=10,top=5)
                    ]))
            elif typ == DirectoryPath:
                text_field = ft.TextField(
                    label=attribute_name.removeprefix("user_"),
                    border_color=ft.Colors.BLUE_ACCENT,
                    value=format_directory_path(value.path, 50),
                    height=60,
                    read_only=True,
                    disabled=True
                )
                dir_picker = ft.FilePicker(
                    on_result=lambda a, attr_name=attribute_name, text=text_field: self.on_select_dir(a, attr_name,
                                                                                                       text))
                self.pipeline_gui.page.overlay.extend([dir_picker])
                items.append(ft.Stack([text_field, ft.Container(
                        content=ft.IconButton(
                            icon=ft.Icons.FOLDER_OPEN,
                            tooltip="Choose directory",
                            on_click=lambda e: dir_picker.get_directory_path(),
                        ),
                        alignment=ft.alignment.top_right,right=10,top=5
                    )
                ]))
            else:
                raise ValueError(f"Unsupported 'user_' attribute file type: {typ}")
        return items

    def on_select_file(self,e,attr_name,text):
        """
        Handles if a file is selected.
        """
        if e.files is not None:
            current_file_path = getattr(self.module, attr_name)
            current_file_path.path = e.files[0].path
            text.value = format_directory_path(e.files[0].path,50)
            text.update()
            self.pipeline_gui.page.update()
            self.pipeline_gui.pipeline.event_manager.notify(OnPipelineChangeEvent("user_attr_change"))

    def on_select_dir(self,e,attr_name,text):
        """
        Handles if a directory is selected.
        """
        if e.path is not None:
            setattr(self.module, attr_name, FilePath(e.path))
            text.value = format_directory_path(e.path,50)
            text.update()
            self.pipeline_gui.page.update()
            self.pipeline_gui.pipeline.event_manager.notify(OnPipelineChangeEvent("user_attr_change"))

    def on_change(self,e,attr_name,reference,typ:type):
        """
        Handles changes to the attribute for different types.
        """
        attribute_name_without_prefix = attr_name.removeprefix("user_")
        if str(e.control.value) == "":
            self.pipeline_gui.page.open(
                ft.SnackBar(
                    ft.Text(f"{attribute_name_without_prefix} must not be empty!",
                            color=ft.Colors.WHITE),
                    bgcolor=ft.Colors.RED))
            e.control.value = str(getattr(self.module, attr_name))
            self.pipeline_gui.page.update()
            return
        try:
            setattr(self.module, attr_name, typ(e.control.value))
            self.pipeline_gui.page.update()
            self.pipeline_gui.pipeline.event_manager.notify(OnPipelineChangeEvent("user_attr_change"))
        except ValueError:
            self.pipeline_gui.page.open(ft.SnackBar(ft.Text(f"{attribute_name_without_prefix} only allows {typ.__name__}'s.",color=ft.Colors.WHITE),bgcolor=ft.Colors.RED))
            reference.current.value = str(getattr(self.module, attr_name))
            self.pipeline_gui.page.update()

    def update_bool(self,e,attr_name):
        """
        Handels changes to the attribute for booleans.
        """
        if int(e.data) == 1:
            setattr(self.module, attr_name, True)
        else:
            setattr(self.module, attr_name, False)
        getattr(self.module, "on_change_" + attr_name)()
        self.pipeline_gui.page.update()
        self.pipeline_gui.pipeline.event_manager.notify(OnPipelineChangeEvent("user_attr_change"))

    def open_options(self,e):
        """
        Open options overlay of the module.
        """
        self.page_overlay.open()
        self.options_button.icon_color = ft.Colors.BLACK38
        self.options_button.update()

    def close_options(self,e):
        """
        Called when the overlay gets dismissed.
        """
        self.page_overlay.close()
        self.module.on_settings_dismiss() if self.module.on_settings_dismiss is not None else None
        self.options_button.icon_color = ft.Colors.WHITE60
        self.options_button.update()

    def copy_module(self):
        """
        Called when the copy button is clicked.
        """
        self.copy_button.icon_color = ft.Colors.BLACK38
        self.copy_button.update()
        copy_dict = self.to_dict()
        self.pipeline_gui.add_module(ModuleType(type(self.module)),x=self.left+20,y=self.top+20,module_dict=copy_dict)
        self.copy_button.icon_color = ft.Colors.WHITE60
        self.copy_button.update()

    def to_dict(self):
        """
        Translates the module into a dictionary.
        """
        user_attributes: List[Dict[str, Any]] = []
        for attr_name in self.module.get_user_attributes:
            value = getattr(self.module,attr_name)
            value_type = type(value).__name__
            if isinstance(value,FilePath) or isinstance(value,DirectoryPath):
                value = value.path
            user_attributes.append({"name": attr_name,"value": value,"attr_type":value_type})
        return {
            "module_id": self.module_id,
            "module_name": self.module.gui_config().name,
            "position": {"x":self.left, "y":self.top},
            "user_attributes": user_attributes,
        }

    def update_user_attr(self,module_dict:dict):
        """
        Update user_attributes with a module dict.
        """
        for attr in module_dict.get("user_attributes", []):
            user_attributes = self.module.get_user_attributes
            if attr["name"] in user_attributes:
                if attr["attr_type"] == FilePath.__name__ or attr["attr_type"] == DirectoryPath.__name__:
                    getattr(self.module, attr["name"]).path = attr["value"]
                else:
                    setattr(self.module, attr["name"], attr["value"])




from abc import ABC, abstractmethod
from collections import deque
from enum import Enum
from typing import Callable
from typing import Dict,List

import flet as ft

from cellsepi.backend.main_window.expert_mode.event_manager import EventManager

class FilePath:
    """
    Type to specify FilePath's
    """
    def __init__(self, path: str = "", suffix: List[str]=None):
        self.path = path
        self.suffix = suffix


class DirectoryPath:
    """
    Type to specify DirectoryPath's
    """
    def __init__(self, path: str = ""):
        self.path = path


class Port:
    """
    Ports defines an input or output of a module.
    Ports with the same names in different modules are considered as the same type of ports
    and their data can be transferred with pipes to each other.
    """
    def __init__(self, name: str, data_type: type, opt: bool = False):
        self.name = name
        self.data_type = data_type #type
        self.opt = opt #optional
        self._data = None

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        """
        Raises:
            TypeError: If the data type is not the required type.
        """
        if isinstance(value, self.data_type) or value is None:
            self._data = value
        else:
            raise TypeError(f"Expected data of type {self.data_type}, got {type(value)}!")

    def __str__(self):
        return f"port_name: '{self.name}', port_data_type: '{self.data_type.__name__}', opt: {self.opt}, data: {self.data}"

class InputPort(Port):
    """
    InputPorts defines an input of a module.
    Ports with the same names in different modules are considered as the same type of ports
    and their data can be transferred with pipes to each other.
    """
    def __init__(self, name: str, data_type: type, opt: bool = False):
        super().__init__(name, data_type, opt)

class OutputPort(Port):
    """
    OutputPorts defines an output of a module.
    Ports with the same names in different modules are considered as the same type of ports
    and their data can be transferred with pipes to each other.
    """
    def __init__(self, name: str, data_type: type):
        super().__init__(name, data_type)

class Categories(Enum):
    """
    Categories of the different modules, each with its own color.
    It is forbidden to use red or green, as they are reserved to indicate specific statuses.
    """
    INPUTS = ft.Colors.ORANGE
    OUTPUTS = ft.Colors.LIGHT_BLUE
    FILTERS = ft.Colors.PURPLE_ACCENT
    MANUAL = ft.Colors.PINK
    SEGMENTATION = ft.Colors.AMBER_ACCENT

class ModuleGuiConfig:
    """
    Stores configuration information for a module's GUI representation.
    """
    def __init__(self, name: str, category: Categories, description:str = None):
        self.name = name
        self.category = category
        self.description = description

class IdNumberManager:
    """
    Manages the module ID's so every module has a unique ID.
    """
    def __init__(self):
        self._occupied_id_numbers = set()
        self._next_id_number = 0

    def get_id_number(self) -> int:
        """
        Gets the next free id number.
        """
        while self._next_id_number in self._occupied_id_numbers:
            self._next_id_number += 1
        id_number = self._next_id_number
        self._occupied_id_numbers.add(id_number)
        self._next_id_number += 1
        return id_number

    def occupy_id_number(self, id_number: int):
        """
        Occupies the given id number so no other module can get it.
        """
        if id_number in self._occupied_id_numbers:
            raise ValueError(f"Number {id_number} already occupied!")
        self._occupied_id_numbers.add(id_number)
        if id_number ==  self._next_id_number:
            self._next_id_number = id_number + 1

    def free_id_number(self, id_number: int) -> None:
        """
        Frees the id number given.
        """
        if id_number in self._occupied_id_numbers:
            self._occupied_id_numbers.discard(id_number)
            if self._next_id_number > id_number >= 0:
                self._next_id_number = id_number


class Module(ABC):
    """
    Modules are independent processes within the pipeline that perform a specific task.
    The modules should be designed to function independently of other modules,
    as long as the correct inputs are provided.

    You can specify user attributes with 'user_' as prefix.
    With these automatic overlay gets created if settings is None.
    """
    @abstractmethod
    def __init__(self,module_id: str):
        self.module_id:str = module_id
        self.event_manager: EventManager | None = None
        self.inputs: Dict[str,InputPort] = {}
        self.outputs: Dict[str,OutputPort] = {}
        self._settings: ft.Stack | None = None
        self._on_settings_dismiss: Callable[[], None] | None = lambda : None
        """
        User-defined attributes convention:        
        - Add custom attributes by prefixing them with 'user_'.
          Example: user_example: str = "Example"
        - Always initialize user attributes with a non-empty value.
        - Supported types: int, float, str, bool, FilePath, DirectoryPath.        
        - User attributes are also saved when the pipeline is saved.
        - When `_settings` is None, GUI elements are automatically generated.
              - For attributes of type int, float, or str, a corresponding reference
                to the GUI element is also automatically generated, named with the
                prefix 'ref_'. Example: ref_user_example
              - For attributes of type bool, an on_change event handler is automatically
                generated. Its name is built with the prefix 'on_change_' followed by the 
                attribute name. Example: on_change_user_example
        """


    @classmethod
    def get_id(cls) -> str:
        """
        Returns the module ID.
        """
        if not hasattr(cls, "_id_number_manager"):
            cls._id_number_manager = IdNumberManager()
        return cls.gui_config().name + str(cls._id_number_manager.get_id_number())

    @classmethod
    def occupy_id_number(cls,id_number: int):
        """
        Occupies the given ID number in the id number manager.
        """
        if not hasattr(cls, "_id_number_manager"):
            cls._id_number_manager = IdNumberManager()
        cls._id_number_manager.occupy_id_number(id_number)

    @classmethod
    def free_id_number(cls, id_number: int):
        """
        Gives the given id number free for other modules.
        """
        if hasattr(cls, "_id_number_manager"):
            cls._id_number_manager.free_id_number(id_number)

    @classmethod
    def destroy_id_number_manager(cls):
        """
        Destroys the id number manager.
        """
        if hasattr(cls, "_id_number_manager"):
            del cls._id_number_manager

    @classmethod
    def gui_config(cls) -> ModuleGuiConfig:
        """
        Returns the module gui config which has the name of the module its category and a description.
        """
        return cls._gui_config

    def occupy(self):
        """
        Occupies the currently module_id the module has in the id number manager.
        """
        id_number = self.module_id.removeprefix(self.gui_config().name)
        if id_number != "":
            number = int(id_number)
            self.occupy_id_number(number)

    def get_id_number(self)-> int:
        """
        Gets the module ID's number.
        """
        id_number = self.module_id.removeprefix(self.gui_config().name)
        return int(id_number)

    def destroy(self):
        """
        Module gets distroyed so free the id_number for other modules.
        """
        id_number = self.module_id.removeprefix(self.gui_config().name)
        if id_number != "":
            number = int(id_number)
            self.free_id_number(number)

    def get_mandatory_inputs(self) -> List[str]:
        """
        Returns the list of names of input ports that are required by the module.
        """
        mandatory_inputs = []
        for port in self.inputs.values():
            if not port.opt:
                mandatory_inputs.append(port.name)
        return mandatory_inputs

    @property
    def settings(self) -> ft.Stack|None:
        """
        The settings overlay of the module in the gui.
        If it is None it gets generated automatically if the modules has user_attributes.
        """
        return self._settings

    def finished(self):
        """
        Gets executed when the module is complety finished include possible pausing.
        """
        return

    @property
    def on_settings_dismiss(self) -> Callable[[], None]:
        """
        The function called when the settings get dismiss.
        """
        return self._on_settings_dismiss

    @property
    def get_user_attributes(self) -> list[str]:
        """
        Returns the list of attributes of the module's user attributes.
        """
        return [k for k in self.__dict__ if k.startswith("user_")]

    @abstractmethod
    def run(self) -> bool: #pragma: no cover
        """
        Returns True if the pipeline should pause.
        """
        pass

    def __str__(self):
        return f"module_id: '{self.module_id}', category: '{self.gui_config().category}', module_name: {self.gui_config().name}, inputs: {self.inputs}, outputs: {self.outputs}, user_attributes: {self.get_user_attributes}"





import ast
import json
import os
import shutil
import time
from threading import Lock

class DeletionForbidden(Exception):
    """
    Custom exception raised when a profile deletion is not allowed.

    This exception is used to indicate that a profile cannot be deleted,
    either because it is the last remaining profile or because the specified
    profile name does not exist.
    """
    pass

def load_config(file_directory):
    """
    Return the current config.json.

    It tries 5 times to read the file,
    before resetting the file to the default config.

    Args:
        file_directory (str): The directory of the file.

    Returns:
        create_default_config() (dict): If the current config.json is corrupted or not existing.
        json.load(config.json) (dict): If the current config.json is readable.
    """
    for _ in range(5):
        try:
            with open(file_directory, 'r') as file:
                return json.load(file)
        except (json.JSONDecodeError, FileNotFoundError):
            time.sleep(0.1)

    return reset_config(file_directory)


def reset_config(file_directory):
    """
    Creates a new default config and tries to save it.
    Args:
        file_directory (str): The directory of the file.

    Returns:
        create_default_config() (dict)
    """
    config = create_default_config()
    try:
        with open(file_directory, 'w') as file:
            json.dump(config, file, indent=4)
    except Exception as e:
        print(f"Error while saving config: {e}")
    return config

def create_default_config():
    """
    Create the default config.

    Returns:
        default_config (dict): the default config.json.
    """

    return {
        "Profiles": {
            "Lif": {
                "bf_channel": "2",
                "mask_suffix": "_seg",
                "channel_prefix": "c",
                "diameter": 125.0
                },
            "Tif": {
                "bf_channel": "1",
                "mask_suffix": "_seg",
                "channel_prefix": "c",
                "diameter": 250.0
                }
        },
        "Selected Profile": {
            "name": "Lif"
        },
        "Colors":{
            "mask": "(255, 0, 0)",
            "outline": "(0, 255, 0)",
        },
        "States": {
            "auto_button": False,
            "lif_slider": True,
        }
    }

class ConfigFile:
    """
    Manages the application's configuration file (config.json).

    This class provides methods to read, write, and modify the configuration
    file. It ensures proper validation of profile attributes and manages
    selected profiles.

    Attributes:
        project_root (str): The root directory of the project.
        file_directory (str): The full path to the configuration file.
        config (dict): The loaded configuration data.
        config_lock (Lock): a lock to make writing the config file save.
    """
    def __init__(self,filename="config.json"):
        self.project_root = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
        self.file_directory = os.path.join(self.project_root, filename)
        self.config = load_config(self.file_directory)
        self.config_lock = Lock()

    def save_config(self):
        """
        Saves the current configuration to the config.json file.

        This method writes the current state of the "config" attribute
        to the "config.json" file.
        """
        with self.config_lock:
            try:
                with open(self.file_directory, 'w') as file:
                    json.dump(self.config, file, indent=4)

            except Exception as e:
                print(f"Error while saving config: {e}")


    def add_profile(self, name:str, bf_channel: int, mask_suffix:str, channel_prefix:str, diameter: float):
        """
        Adds a new profile to the config.

        Args:
            name (str): Name of the new profile.
            bf_channel (int): Bright-field channel for the profile.
            mask_suffix (str): Mask suffix for the profile.
            channel_prefix (str): Channel prefix for the profile.
            diameter (float): Diameter value for the profile.

        Invalid:
            - Strings ("mask_suffix", "channel_prefix" and "bf_channel") must not be empty ("").
            - Numeric values ("diameter") must be greater than 0.

        Raises:
            ValueError: If any provided parameter is invalid.

        Returns:
            False: If the profile name is already taken.
            True: If the profile is successfully added.
        """
        if not all([name, mask_suffix, channel_prefix, bf_channel]):
            raise ValueError("Name, mask_suffix, bf_channel, and channel_prefix must not be empty.")
        if diameter <= 0 :
            raise ValueError("diameter must be greater than 0.")
        if not name in self.config['Profiles']:
            self.config["Profiles"][name] = {
                "bf_channel": bf_channel,
                "mask_suffix": mask_suffix,
                "channel_prefix": channel_prefix,
                "diameter": diameter
            }
            self.save_config()
            return True
        else:
            return False

    def update_profile(self, name: str, bf_channel: str = None, mask_suffix: str = None,
                       channel_prefix: str = None, diameter: float = None):
        """
        Updates the attributes of a profile.

        Args:
            name (str): Name of the profile to update.
            bf_channel (str, optional): New bright-field channel.
            mask_suffix (str, optional): New mask suffix.
            channel_prefix (str, optional): New channel prefix.
            diameter (float, optional): New diameter value.

        Invalid:
            - Strings ("mask_suffix" and "channel_prefix") must not be empty ("").
            - Numeric values ("bf_channel" and "diameter") must be greater than 0.

        Raises:
            ValueError: If any provided parameter is invalid.
        """
        if name in self.config["Profiles"]:
            if bf_channel is not None:
                if not bf_channel:
                    raise ValueError("bf_channel must not be empty.")
                self.config["Profiles"][name]["bf_channel"] = bf_channel
            if mask_suffix is not None:
                if not mask_suffix:
                    raise ValueError("mask_suffix must not be empty.")
                self.config["Profiles"][name]["mask_suffix"] = mask_suffix
            if channel_prefix is not None:
                if not channel_prefix:
                    raise ValueError("channel_prefix must not be empty.")
                self.config["Profiles"][name]["channel_prefix"] = channel_prefix
            if diameter is not None:
                if diameter <= 0:
                    raise ValueError("diameter must be greater than 0.")
                self.config["Profiles"][name]["diameter"] = diameter
            self.save_config()

    def rename_profile(self,old_name: str,new_name: str):
        """
        Renames the old name to the new name.

        Args:
            old_name (str): The current/old name of the profile.
            new_name (str): The new name for the profile.

        Raises:
            ValueError: If any provided parameter is empty ("").

        Returns:
            False: If the new profile name is already taken or the old name does not exist.
            True: If the new profile is successfully renamed.
        """
        if not all([old_name,new_name]):
            raise ValueError("old_name, new_name must not be empty.")
        elif old_name == new_name:
            return True
        elif old_name in self.config["Profiles"] and new_name not in self.config["Profiles"]:
            self.config["Profiles"][new_name] = self.config["Profiles"].pop(old_name)
            self.save_config()
            if old_name == self.get_selected_profile_name():
                self.select_profile(new_name)
            return True
        else:
            return False

    def get_profile(self, name):
        """
        Gets a profile by name.

        Args:
            name (str): The name of the profile.

        Returns:
            profile (dict): A dictionary containing the profile's attributes.
        """
        if name in self.config["Profiles"]:
            return self.config["Profiles"][name]

    def delete_profile(self, name: str):
        """
        Deletes a profile by name.

        Args:
            name (str): The name of the profile.

        Raises:
             DeletionForbidden: If the profile count is equal to 1 or the profile does not exist.
        """
        if name in self.config["Profiles"] and len(self.config["Profiles"]) > 1:
            del self.config["Profiles"][name]
            if self.config["Selected Profile"]["name"] == name:
                first_key = next(iter(self.config["Profiles"]))
                self.config["Selected Profile"]["name"] = first_key
            self.save_config()
        else:
            raise DeletionForbidden


    def select_profile(self,name: str):
        """
        Selects a profile by name.

        Args:
            name (str): The name of the profile to select.

        Raises:
              ValueError: If the profile is empty ("").
        """
        if not name:
            raise ValueError("name must not be empty.")
        elif name in self.config["Profiles"]:
            self.config["Selected Profile"]["name"] = name
            self.save_config()

    def get_selected_profile_name(self):
        """
        Gets the name of the selected profile.

        Returns:
            profile name (str): The name of the selected profile or if the no profile is selected the first profile.
        """
        if self.config["Selected Profile"]["name"] is not None:
            return self.config["Selected Profile"]["name"]
        else:
            first_key = next(iter(self.config["Profiles"]))
            self.select_profile(first_key)
            return first_key

    def name_to_index(self, name: str):
        """
        Converts a profile name to its index.

        Args:
            name (str): The name of the profile.

        Raises:
            ValueError: If the profile does not exist.

        Returns:
            index (int): The index of the profile.
        """
        profiles = list(self.config["Profiles"].keys())
        if name in profiles:
            return profiles.index(name)
        else:
            raise ValueError("Profile with that name does not exists")

    def index_to_name(self, index: int):
        """
        Converts a profile index to its name.

        Args:
            index (int): The index of the profile.

        Raises:
            ValueError: If no profile is at this index.

        Returns:
            profile name (str): The name of the profile.
        """
        profiles = list(self.config["Profiles"].keys())
        if 0 <= index < len(profiles):
            return profiles[index]
        else:
            raise ValueError("Didnt find a profile at this index")

    def is_profile_existing(self, name: str):
        """
        Checks if a profile exists.

        Args:
            name (str): The name of the profile.

        Returns:
            Boolean: True if the profile exists, False otherwise.
        """
        return name in self.config["Profiles"]

    #------------------------------------------
    #getter for the selected profiles Attributes

    def get_selected_profile(self):
        """
        Gets the selected profile.

        Returns:
            profile (dict): A dictionary containing the profile's attributes.
        """
        name = self.get_selected_profile_name()
        return self.config["Profiles"][name]

    def get_bf_channel(self):
        """
        Gets the bright-field channel for the profile.

        Returns:
            bright-field channel (str): The bright-field channel for the profile.
        """
        profile = self.get_selected_profile()
        return profile["bf_channel"]

    def get_mask_suffix(self):
        """
        Gets the mask suffix for the profile.

        Returns:
            mask_suffix (str): The mask suffix for the profile.
        """
        profile = self.get_selected_profile()
        return profile["mask_suffix"]

    def get_channel_prefix(self):
        """
        Gets the channel prefix for the profile.

        Returns:
            channel_prefix (str): The channel prefix for the profile.
        """
        profile = self.get_selected_profile()
        return profile["channel_prefix"]

    def get_diameter(self):
        """
        Gets the diameter for the profile.

        Returns:
            diameter (float): The diameter for the profile.
        """
        profile = self.get_selected_profile()
        return float(profile["diameter"])

    def get_mask_color(self):
        """
        Gets the selected mask color.

        Returns:
            mask_color (tuple): The selected mask color in RGB format.
        """
        return ast.literal_eval(self.config["Colors"]["mask"])
    def get_outline_color(self):
        """
        Gets the selected outline color.

        Returns:
            outline_color (tuple): The selected mask color in RGB format.
        """
        return ast.literal_eval(self.config["Colors"]["outline"])
    def set_mask_color(self, color):
        """
        Sets the selected mask color in the config file.

        Args:
            color (tuple): The selected mask color in RGB format.

        Raises:
            ValueError: If the parameter color is invalid.
        """
        if isinstance(color, tuple) and len(color) == 3:
            self.config["Colors"]["mask"] = f"{color}"
            self.save_config()
        else:
            raise ValueError("Color must be an RGB tuple, e.g., (255, 0, 0)")
    def set_outline_color(self, color):
        """
        Sets the selected outline color in the config file.

        Args:
            color (tuple): The selected outline color in RGB format.
        Raises:
            ValueError: If the parameter color is invalid.
        """
        if isinstance(color, tuple) and len(color) == 3:
            self.config["Colors"]["outline"] = f"{color}"
            self.save_config()
        else:
            raise ValueError("Color must be an RGB tuple, e.g., (0, 255, 0)")
    def get_auto_button(self):
        """
        Gets the last state of the auto bright-field and contrast button.
        Returns:
            bool: The last state of the auto bright-field and contrast button.
        """
        return self.config["States"]["auto_button"]
    def get_lif_slider(self):
        """
        Gets the last state of the lif-slider button.
        Returns:
            bool: The last state of the lif-slider button (True means lif is active).
        """
        return self.config["States"]["lif_slider"]
    def set_auto_button(self, val:bool):
        """
        Sets the state of the auto bright-field and contrast button.
        """
        self.config["States"]["auto_button"] = val
        self.save_config()
    def set_lif_slider(self, val:bool):
        """
        Sets the state of the lif-slider button (True means lif is active).
        """
        self.config["States"]["lif_slider"] = val
        self.save_config()
    #-----------------------------------------------------
    #only for test_config
    def clear_config(self):
        """
        Only for Training
        ___________________________
        Makes a backup of the current config
        and deletes the original config.
        Then it loads the deleted config to trigger the default config to load.
        """
        backup_filepath = os.path.join(self.project_root, "config_backup.json")
        shutil.copy(self.file_directory, backup_filepath)
        open(self.file_directory, 'w').close()
        self.config = load_config(self.file_directory)

    def restore_config(self):
        """
        Only for Training
        ___________________________
        Copy's the backup back into the config.json file and
        refreshes the ConfigFile class with the "new" values.
        Then deletes the backup.
        """
        backup_filepath = os.path.join(self.project_root, "config_backup.json")
        shutil.copy(backup_filepath, self.file_directory)
        self.config = load_config(self.file_directory)
        os.remove(backup_filepath)



import asyncio
import os
import pathlib
import platform
from collections import defaultdict
import concurrent.futures
from json.encoder import INFINITY
from time import time
import flet as ft
import tifffile

from cellsepi.backend.main_window.expert_mode.event_manager import EventManager
from cellsepi.backend.main_window.expert_mode.listener import ProgressEvent
from cellsepi.backend.main_window.expert_mode.pipeline import PipelineRunningException
from cellsepi.frontend.main_window.gui_canvas import update_main_image
from cellsepi.frontend.main_window.gui_fluorescence import fluorescence_button
from cellsepi.backend.main_window.data_util import extract_from_lif_file, copy_files_between_directories, load_directory, transform_image_path, \
    convert_tiffs_to_png_parallel

def format_directory_path(dir_path: str, max_length=30):
    """
    Format the directory so that it can be shown in the card.
    Args:
        dir_path (str): Path to the directory that should be formatted.
        max_length (int): Maximum length of the directory path.
    """
    parts = dir_path.split('/')
    path = dir_path
    if len(dir_path) > max_length:
        if len(parts) > 2:
            path = f".../{parts[len(parts) - 2]}/{parts[len(parts) - 1]}"
        else:
            return f"...{path[len(parts) - (max_length - 3):]}"

    if len(path) > max_length:
        path = f"...{path[len(parts) - (max_length - 3):]}"  # 3 für '...'

    return path

def copy_to_clipboard(page,value: str,name: str):
    """
    Adds the value in to the clipboard and opens the snack_bar and say that it has been copied.
    Args:
        page: ft.Page object.
        value (str): Value to add to the clipboard.
        name (str): Name of the thing that got copied.
    """
    page.set_clipboard(value)
    page.open(ft.SnackBar(ft.Text(f"{name} copied to clipboard!")))
    page.update()


def get_image(linux_or_3d: bool, src):
    """
    Adjust the method of reading the image, depending on whether the system is Linux and if the images are 3d or not (src_Base64 or src).
    Args:
        linux_or_3d (bool): Whether the system is Linux or the files are 3d.
        src (str): The path of the image to load.
    Returns:
          ft.Image: the image at the src path.
    """
    if linux_or_3d:
        return ft.Image(
            src_base64=src,
            height=150,
            width=150,
            fit=ft.ImageFit.CONTAIN
        )
    else:
        return ft.Image(
            src=src,
            height=150,
            width=150,
            fit=ft.ImageFit.CONTAIN
        )


class DirectoryCard(ft.Card):
    """
    Handles the directory card with all event handlers.
    """
    def __init__(self, gui = None):
        if gui is not None:
            super().__init__()
            self.gui = gui
            self.count_results_txt = ft.Text(value="Results: 0")
            self.directory_path = ft.Text(value='Directory Path',weight=ft.FontWeight.BOLD)
            self.formatted_path = ft.Text(value=format_directory_path(self.directory_path.value), weight=ft.FontWeight.BOLD)
            self.is_lif = self.gui.csp.config.get_lif_slider()
            if self.is_lif:
                index = 1
            else:
                index = 0
            self.lif_slider = ft.CupertinoSlidingSegmentedButton(
                selected_index=index,
                thumb_color=ft.Colors.BLUE_400,
                on_change=self.update_view,
                padding=ft.padding.symmetric(0, 0),
                controls=[
                    ft.Text("Tif"),
                    ft.Text("Lif")
                ],
            )
            self.image_gallery = ft.ListView()
            self.path_list_tile = self.create_path_list_tile()
            self.get_directory_dialog = None
            self.pick_files_dialog = None
            self.create_handlers()
            self.directory_row = self.create_dir_row()
            self.files_row = self.create_files_row()
            self.lif_slider_blocker = ft.Container(
                width=80,
                height=30,
                bgcolor=ft.Colors.TRANSPARENT,
                on_click= None,
                visible= False,
            )
            self.lif_row = ft.Row([ft.Stack([self.lif_slider,self.lif_slider_blocker]),
                                                   self.directory_row,
                                                   self.files_row
                                                    ], alignment=ft.MainAxisAlignment.SPACE_BETWEEN
                                                  )
            self.content = self.create_directory_container()
            self.output_dir = False
            self.is_supported_lif = True
            self.files_row.visible = self.is_lif
            self.directory_row.visible = not self.is_lif
            self.selected_images_visualise = {}
            self.icon_check = {}
            self.icon_x = {}

    def create_path_list_tile(self):
        return ft.ListTile(leading=ft.Icon(name=ft.Icons.FOLDER_OPEN),
                    title=self.formatted_path,
                    subtitle=self.count_results_txt
                    )

    def update_results_text(self):
        self.count_results_txt.value = f"Results: {len(self.gui.csp.image_paths)}"
        self.count_results_txt.update()

    def get_directory_result(self, e: ft.FilePickerResultEvent):
        """
        Checks if the picked directory or file exists and if it worked updates everything with the new values.
        builds the canvas container for the file results on the right column of the GUI
        """
        if not(e.files is None and e.path is None):
            self.gui.progress_ring.visible = True
            self.image_gallery.controls.clear()
            self.gui.canvas.main_image.content = ft.Image(
                src=r"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAUA\AAAFCAIAAAFe0wxPAAAAAElFTkSuQmCC",
                fit=ft.ImageFit.SCALE_DOWN)
            # the window of the image display is cleared of all content
            self.gui.switch_mask.value = False
            self.gui.canvas.container_mask.visible = False
            self.gui.csp.image_id = None
            self.gui.csp.channel_id = None
            self.gui.open_button.visible = False
            self.gui.drawing_button.disabled = True
            self.gui.start_button.disabled = True
            self.gui.training_environment.start_button.disabled = True
            fluorescence_button.visible = False
            self.gui.progress_bar_text.value = "Waiting for Input"
            self.gui.progress_bar.value = 0
            self.gui.contrast_slider.disabled = True
            self.gui.brightness_slider.disabled = True
            self.gui.csp.current_channel_prefix = self.gui.csp.config.get_channel_prefix()
            self.gui.csp.current_mask_suffix = self.gui.csp.config.get_mask_suffix()
            self.gui.mask.mask_outputs = defaultdict(dict)
            self.gui.contrast_slider.value = 1
            self.gui.brightness_slider.value = 1
            self.gui.diameter_display.opacity = 0.5
            if not platform.system() == "Linux":
                self.gui.page.window.progress_bar = -1
            self.gui.page.update()
            self.gui.queue.put("delete_mask")

            #differentiate between the lif and tiff case, as there are two different file formats
            if self.is_lif:
                #is a file
                path = e.files[0].path
            else:
                #is a directory
                path = e.path
            if path:
                self.directory_path.value = path
                self.select_directory_parallel(path,self.is_lif,self.gui.csp.config.get_channel_prefix())
                self.load_images()
            else:
                self.image_gallery.controls.clear()
                self.image_gallery.update()

            self.formatted_path.value = format_directory_path(self.directory_path.value)
            if self.output_dir or not self.is_supported_lif:
                self.formatted_path.color = ft.Colors.RED
                self.gui.diameter_text.value = 0.0
                self.gui.diameter_display.update()
            else:
                self.gui.diameter_text.value = self.gui.average_diameter.get_avg_diameter()
                self.gui.diameter_display.opacity = 1
                self.gui.diameter_display.update()
                self.formatted_path.color = None
            self.formatted_path.update()

    def select_directory_parallel(self, directory_path ,is_lif:bool,channel_prefix: str ,event_manager: EventManager = None):
        """
            Gets the working directory and copies the images in there.

            Args:
                directory_path (str): the selected directory_path
                is_lif (bool): if the images are lif or tif
                channel_prefix (str): the channel prefix
                event_manager (EventManager): the event manager which is used when the methode gets started as a module.
                """
        is_supported_tif = True
        if event_manager is None:
            self.is_supported_lif = True
        path = pathlib.Path(directory_path)
        # Lif Case
        if is_lif:
            if event_manager is None:
                self.output_dir = False
            working_directory = path.parent / "output/"
            os.makedirs(working_directory, exist_ok=True)
            if path.suffix.lower() == ".lif":
                # Extract from a lif file all the single series images and extract to .tif, .tiff and .npy files into subdirectory
                extract_from_lif_file(lif_path=path, target_dir=working_directory,channel_prefix=channel_prefix,event_manager=event_manager)
            else:
                if event_manager is not None:
                    raise PipelineRunningException("Type Error","Expected .lif!")
                else:
                    self.is_supported_lif = False


        # Tiff Case
        else:
            if path.name == "output":
                if event_manager is not None:
                    raise PipelineRunningException("Directory Error","Directory ’output’ is not supported!")
                else:
                    self.gui.page.open(ft.SnackBar(ft.Text("The directory path output is not allowed!")))
                    self.output_dir = True
                    self.gui.page.update()
                    self.gui.csp.image_paths = {}
                    self.gui.csp.linux_images = {}
                    self.gui.csp.mask_paths = {}
                    self.gui.ready_to_start = False
                    self.gui.progress_ring.visible = False
                    return None
            if event_manager is None:
                self.output_dir = False
            # Copy .tif, .tiff and .npy files into subdirectory
            working_directory = path / "output/"
            os.makedirs(working_directory, exist_ok=True)
            copy_files_between_directories(path, working_directory, file_types=[".tif", ".tiff", ".npy"],event_manager=event_manager)
            tiff_paths = [p for p in working_directory.iterdir() if
                          p.suffix.lower() in [".tif", ".tiff"] and p.is_file()]
            total = len(tiff_paths)
            converted_count = 0
            failed = False

            if event_manager is not None:
                event_manager.notify(ProgressEvent(percent=0, process=f"Convert TIFFs: {converted_count}/{total}"))

            with concurrent.futures.ThreadPoolExecutor() as executor:
                futures = {executor.submit(self.convert_tiffs_to_8_bit, path): path for path in tiff_paths}
                for future in concurrent.futures.as_completed(futures):
                    result = future.result()
                    if not result:
                        failed = True
                    converted_count += 1
                    if event_manager is not None:
                        event_manager.notify(
                            ProgressEvent(percent=int(converted_count / total*100),process=f"Convert Tiff's: {converted_count}/{total}")
                        )

            if failed:
                if event_manager is not None:
                    raise PipelineRunningException("Type Error",
                                                   "The directory contains an unsupported file type. Only 8 or 16 bit .tiff files allowed.")
                else:
                    is_supported_tif = False
            else:
                if event_manager is not None:
                    event_manager.notify(
                        ProgressEvent(percent=100,
                                      process=f"Finished converting Tiff's!")
                    )

        if event_manager is not None:
            return working_directory
        else:
            self.gui.csp.working_directory = working_directory
            self.set_paths(is_supported_tif)
            return None

    def convert_tiffs_to_8_bit(self, path):
        """
        handles when the conversion should happen
        Args:
            path (str): the selected directory_path
        """
        converted=True
        if path.suffix.lower() == ".tif" or path.suffix.lower() == ".tiff":
            if path.is_file():
                converted=transform_image_path(path, path)
        return converted

    def set_paths(self, is_supported_tif):
        """
        Updates the image and mask paths in csp (CellSePi).

        Args:
             is_supported_tif (bool): True if the image types for tif are supported.
        """
        bfc = self.gui.csp.config.get_bf_channel()
        cp = self.gui.csp.config.get_channel_prefix()
        ms = self.gui.csp.config.get_mask_suffix()
        working_directory = self.gui.csp.working_directory

        if not self.is_supported_lif:
            self.gui.ready_to_start = False
            self.gui.page.open(ft.SnackBar(
                ft.Text("The selected file is not supported! Only .lif are supported.")))
            image_paths = {}
            mask_paths = {}
            self.gui.progress_ring.visible = False
            self.gui.page.update()
        else:
            image_paths, mask_paths = load_directory(working_directory, channel_prefix=cp, mask_suffix=ms)
            if len(image_paths) == 0:
                self.gui.ready_to_start = False
                self.gui.page.open(ft.SnackBar(ft.Text("The directory contains no valid files with the current channel prefix!")))
                self.gui.page.update()
                self.count_results_txt.color = ft.Colors.RED
                self.gui.progress_ring.visible = False
                if not self.is_lif:
                    os.rmdir(self.gui.csp.working_directory)
            elif not is_supported_tif:
                self.gui.ready_to_start = False
                self.gui.page.open(ft.SnackBar(ft.Text("The directory contains an unsupported file type. Only 8 or 16 bit .tiff files allowed.")))
                self.count_results_txt.color = ft.Colors.RED
                self.gui.progress_ring.visible = False
                self.gui.page.update()
                image_paths = {}
                mask_paths = {}
            else:
                self.count_results_txt.color = None
                self.gui.training_environment.start_button.disabled = False
                if self.gui.csp.model_path is not None:
                    self.gui.progress_bar_text.value = "Ready to Start"
                    self.gui.start_button.disabled = False
                self.gui.ready_to_start = True

        self.gui.csp.image_paths = image_paths
        self.gui.csp.mask_paths = mask_paths


    def load_images(self):
        """
        Load images to gallery in order and with names.
        """

        self.page.run_task(self.check_masks)

        self.gui.page.update()

        src = self.gui.csp.image_paths

        is_3d = any(
            tifffile.imread(channel_path).ndim == 3
            for outer_dict in src.values()
            for channel_path in outer_dict.values()
        )

        if platform.system() == "Linux" or is_3d:
            self.gui.csp.linux_images = convert_tiffs_to_png_parallel(self.gui.csp.image_paths)
            self.gui.csp.linux_or_3d = True
            src = self.gui.csp.linux_images

        self.selected_images_visualise = {}
        # Display groups with side-by-side images for linux_or_3d
        for image_id in src:
            cur_image_paths = src[image_id]
            self.selected_images_visualise[image_id] = {}
            for channel_id in cur_image_paths:
                self.selected_images_visualise[image_id][channel_id] = ft.Container(
                    width=154,
                    height=154,
                    border=ft.border.all(4, ft.Colors.ORANGE_700),
                    alignment=ft.alignment.center,
                    visible=False,
                    padding=5
                )
            group_row = ft.Row(
                [
                    ft.Column(
                    [
                            ft.GestureDetector(
                                content=ft.Container(ft.Stack([get_image(self.gui.csp.linux_or_3d,
                                                                         cur_image_paths[channel_id]), self.selected_images_visualise[image_id][channel_id]]), width=156, height=156),
                                on_tap=lambda e, img_id=image_id, c_id=channel_id: update_main_image(img_id, c_id,
                                                                                                     self.gui),
                            ),
                            ft.Text(channel_id, size=10, text_align=ft.TextAlign.CENTER),
                        ],
                        alignment=ft.MainAxisAlignment.CENTER,
                        horizontal_alignment=ft.CrossAxisAlignment.CENTER,
                        spacing=5
                    )
                    for channel_id in cur_image_paths
                ],
                alignment=ft.MainAxisAlignment.START,
                spacing=10,
                scroll=ft.ScrollMode.AUTO,
            )
            self.icon_check[image_id] = ft.Icon(ft.Icons.CHECK, color=ft.Colors.GREEN, size=17, visible=False,
                                                tooltip="Mask is available")
            self.icon_x[image_id] = ft.Icon(ft.Icons.CLOSE, size=17, visible=True, tooltip="Mask not available")
            self.update_mask_check(image_id)
            self.image_gallery.controls.append(ft.Column([ft.Row(
            [ft.Text(f"{image_id}", weight=ft.FontWeight.BOLD, text_align=ft.TextAlign.CENTER), self.icon_check[image_id], self.icon_x[image_id]], spacing=2),
                                                      group_row], spacing=10, alignment=ft.MainAxisAlignment.CENTER))

        self.gui.progress_ring.visible = False
        self.gui.progress_ring.update()
        self.image_gallery.update()
        self.update_results_text()

    def update_mask_check(self, image_id):
        """
        Updates the symbol next to series number of image to a check or x, depending on if the corresponding image is available.
        Args:
            image_id: the id of the image to check mask availability
        """
        if self.gui.csp.mask_paths is not None and image_id in self.gui.csp.mask_paths and self.gui.csp.config.get_bf_channel() in self.gui.csp.mask_paths[image_id]:
            self.icon_check[image_id].visible = True
            self.icon_x[image_id].visible = False
        else:
            self.icon_check[image_id].visible = False
            self.icon_x[image_id].visible = True
        self.image_gallery.update()

    def update_all_masks_check(self):
        """
        Updates the symbol next to series number of image for every image_id in mask_paths.
        """
        if self.gui.csp.image_paths is not None:
            for image_id in self.gui.csp.image_paths:
                self.update_mask_check(image_id)

    def create_dir_row(self):
        """
        Creates the row for directory picking.
        """
        return ft.Row(
        [
            ft.ElevatedButton(
                "Open Directory",
                icon=ft.Icons.FOLDER_OPEN,
                on_click=lambda _: self.get_directory_dialog.get_directory_path(),
                disabled=self.gui.page.web,
            ),
        ], alignment=ft.MainAxisAlignment.START  # Change alignment to extend fully to the left
    )
    def create_files_row(self):
        """
        Creates the row for file picking.
        """
        return ft.Row(
        [
            ft.ElevatedButton(
                "Pick File",
                icon=ft.Icons.UPLOAD_FILE,
                on_click=lambda _: self.pick_files_dialog.pick_files(allow_multiple=False),
            )
        ], alignment=ft.MainAxisAlignment.START  # Change alignment to extend fully to the left
    )
    def create_handlers(self):
        """
        Creates the handlers.
        """
        self.get_directory_dialog = ft.FilePicker(on_result=lambda e: self.get_directory_result(e))
        self.pick_files_dialog = ft.FilePicker(on_result=lambda e: self.get_directory_result(e))
        #add the handlers to the page
        self.gui.page.overlay.extend([self.pick_files_dialog, self.get_directory_dialog])

    def update_view(self,e):
        """
        Changes the visibility of the directory/file picking.
        """
        if int(e.data) == 1:
            self.is_lif = True
            self.gui.csp.config.set_lif_slider(True)
            self.files_row.visible = True
            self.directory_row.visible = False
        else:
            self.is_lif = False
            self.gui.csp.config.set_lif_slider(False)
            self.files_row.visible = False
            self.directory_row.visible = True

        self.gui.page.update()



    def create_directory_container(self):
        return ft.Container(
                content=ft.Stack(
                    [
                        ft.Container(
                            content=ft.Column(
                                [
                                    self.path_list_tile,
                                    self.lif_row
                                ]
                            )
                        ),
                        ft.Container(
                            content=ft.Container(
                                content=ft.IconButton(
                                    icon=ft.Icons.COPY,
                                    tooltip="Copy to clipboard",
                                    on_click=lambda e: copy_to_clipboard(self.gui.page,self.gui.directory.directory_path.value,"Directory path")
                                ),
                                alignment=ft.alignment.top_right,
                            )
                        )
                    ]

                ),
                padding=10,
            )

    def disable_path_choosing(self):
        """
        Disables everything related with path choosing.
        """
        self.path_list_tile.disabled = True
        self.lif_row.disabled = True
        self.toggle_slider_state(self.lif_slider,disabled=True)

        self.gui.page.update()

    def enable_path_choosing(self):
        """
        Activates everything related with path choosing.
        """
        self.path_list_tile.disabled = False
        self.lif_row.disabled = False
        self.toggle_slider_state(self.lif_slider,disabled=False)
        self.gui.page.update()

    def toggle_slider_state(self,slider, disabled):
        """
        Toggles slider state if it is active or not.

        Args:
            slider: Slider object.
            disabled: Boolean if the slider should be disabled.
        """
        if disabled:
            slider.on_change = None
            slider.thumb_color = ft.Colors.GREY_400
            self.lif_slider_blocker.visible = True
            for control in slider.controls:
                control.color = ft.Colors.GREY_700
        else:
            slider.on_change = self.update_view
            slider.thumb_color = ft.Colors.BLUE_400
            self.lif_slider_blocker.visible = False
            for control in slider.controls:
                control.color = None

    async def check_masks(self):
        """
        Check if all masks are present (non-blocking).
        """
        if self.gui.csp.mask_paths is not None:
            bfc = self.gui.csp.config.get_bf_channel()

            loop = asyncio.get_event_loop()
            all_mask_present = await loop.run_in_executor(
                None,
                lambda: all(
                    image_id in self.gui.csp.mask_paths and bfc in self.gui.csp.mask_paths[image_id]
                    for image_id in self.gui.csp.image_paths
                )
            )

            if all_mask_present and self.gui.csp.image_paths is not None and len(self.gui.csp.image_paths) != 0 :
                fluorescence_button.visible = True
                fluorescence_button.update()
            else:
                fluorescence_button.visible = False
                fluorescence_button.update()




import flet as ft
from cellsepi.frontend.main_window.gui_canvas import update_main_image
from cellsepi.backend.main_window.config_file import ConfigFile, create_default_config
from cellsepi.frontend.main_window.gui_page_overlay import PageOverlay


class GUIConfig:
    """
    Manages the GUI config and its elements.

    Attributes:
        config_class (ConfigFile): The instance of the ConfigFile class used to read and update profile data.
        page (Page): The page instance to display GUI elements.
        name_items (List): A list of buttons and text fields for editing or selecting profiles.
        profile_chooser_overlay (PageOverlay): The overlay shown when the profile name button is clicked.
        profile_ref (Ref): Reference to the currently selected profile name button.
        txt_bf_ref (Ref): Reference to the bright-field channel text field.
        txt_ms_ref (Ref): Reference to the mask suffix text field.
        txt_cp_ref (Ref): Reference to the channel prefix text field.
        txt_d_ref (Ref): Reference to the diameter text field.
    """
    def __init__(self,gui):
        """
        Initializes the GUIConfig instance.

        Args:
            gui (GUI): The GUI instance containing ConfigFile and Page references.
        """
        self.config_class: ConfigFile = gui.csp.config
        self.gui = gui
        self.page = gui.page
        self.name_items = self.create_name_items_profiles()
        self.profile_chooser_overlay = self.create_profile_overlay()
        #--------------------------------
        #creates attributes to reference to the text boxes to changes their values or color etc...
        self.profile_ref = ft.Ref[ft.Text]()
        self.txt_bf_ref = ft.Ref[ft.Text]()
        self.txt_ms_ref = ft.Ref[ft.Text]()
        self.txt_cp_ref = ft.Ref[ft.Text]()
        self.txt_d_ref = ft.Ref[ft.Text]()

    def create_profile_overlay(self):
        """
        Creates the overlay for selecting or editing profiles.

        The overlay is displayed when clicking the profile name button.
        It allows users to view, edit, or delete profiles in a scrollable list.

        Returns:
            overlay (CupertinoBottomSheet): The configured overlay for profile selection and editing.
        """
        return PageOverlay(self.page,
            content=ft.Stack([ft.Row([ft.Column(
                [ft.Card(ft.Container(ft.ListView(
                    controls=self.create_list_items(),
                    height=self.calc_height(),
                    width=350,
                    spacing=10,
                    padding=10,
                ),padding=15),height=self.calc_height()+60)],
                alignment=ft.MainAxisAlignment.CENTER,
            )],alignment=ft.MainAxisAlignment.CENTER),]),
            on_dismiss=lambda e: self.update_overlay(),
        )

    def calc_height(self):
        """
        Calculates the height of the profile list in the overlay.

        Returns:
            int: The height of the list. If the profile count exceeds 8, the height is capped at 400.
        """
        if len(self.config_class.config["Profiles"]) < 9:
            return 49 * len(self.config_class.config["Profiles"])
        else:
            return 400

    def text_field_activate(self,e, idx):
        """
        Toggles the activation of a profile's text field for editing.

        Ensures only one text field is active at a time, making the corresponding
        button invisible.

        Args:
            e (Event): The triggering event.
            idx (int): The index of the profile to activate or deactivate.
        """
        if not self.name_items[idx]["textfield"].visible:
            self.name_items[idx]["textfield"].visible = True
            self.name_items[idx]["button"].visible = False
            for i in range(len(self.config_class.config["Profiles"])):
                if not i ==  idx:
                    self.name_items[i]["textfield"].visible = False
                    self.name_items[i]["button"].visible = True
            self.page.update()
        else:
            self.name_items[idx]["textfield"].visible = False
            self.name_items[idx]["button"].visible = True
            self.page.update()

    def text_field_written(self,e, idx):
        """
        Handles the event when a text field is updated and attempts to rename
        the associated profile.

        Args:
            e (Event): The text_field update event.
            idx (int): The index of the profile try to rename.

        Behavior:
        -   Displays an error message if the name is invalid or taken.
        -   Updates the user interface on successful renaming.
        """
        try:
            renamed = self.config_class.rename_profile(self.config_class.index_to_name(idx), e.control.value)
            if not renamed:
                self.page.open(ft.SnackBar(
                    ft.Text("The name is already taken!")))
                self.page.update()
            else:
                self.name_items[idx]["textfield"].visible = False
                self.name_items[idx]["button"].visible = True
                self.name_items[idx]["textfield"].color = None
                self.profile_ref.current.value = self.config_class.get_selected_profile_name()
                self.update_overlay()
                self.page.update()
        except ValueError:
            self.page.open(ft.SnackBar(
                ft.Text("The name must be not empty!")))
            self.page.update()

    def create_name_items_profiles(self):
        """
        Creates TextField and Button for every profile and returns the set as a list.

        Returns:
            List[Dict]: A list of dictionaries, each representing a set of profile item.
        """
        return [
            {
                "textfield": ft.TextField(
                    value=profile,
                    width=200,
                    on_blur=lambda e,i=i:self.text_field_written(e,i),
                    visible = False,
                    border_color=ft.Colors.BLUE_ACCENT,
                ),
                "button": ft.TextButton(
                    content=ft.Text(profile, size=20),
                    on_click=lambda e, i=i: self.selected_profile_changed(e, i,True),
                    width=200,
                    visible = True
                )
            }
            for i, profile in enumerate(self.config_class.config["Profiles"])
        ]

    def selected_profile_changed(self,e, index, called_by_overlay: bool = None):
        """
        Handles the event when a profile is selected, updates the profile attributes,
        and closes the profile selection overlay.

        Args:
            e (Event): The event representing the user's profile selection.
            index (int): The index of the selected profile.
            called_by_overlay (bool): Whether the overlay is called by overlay.

        Behavior:
        -   Selects the profile based on the provided index.
        -   Updates the UI elements with the corresponding profile attributes (e.g.,
            channel, mask suffix, diameter, etc.).
        -   resets the colors of the input fields to their default state, clearing any error indications (such as red color).
        -   Closes the profile chooser overlay and updates the page.
        """
        self.config_class.select_profile(self.config_class.index_to_name(index))
        self.profile_ref.current.value = self.config_class.get_selected_profile_name()
        self.txt_bf_ref.current.value = self.config_class.get_bf_channel()
        self.txt_ms_ref.current.value = self.config_class.get_mask_suffix()
        self.txt_cp_ref.current.value = self.config_class.get_channel_prefix()
        self.txt_d_ref.current.value = self.config_class.get_diameter()
        self.txt_bf_ref.current.color = None
        self.txt_ms_ref.current.color = None
        self.txt_cp_ref.current.color = None
        self.txt_d_ref.current.color = None
        if called_by_overlay:
            self.profile_chooser_overlay.close()
        self.page.update()

    def remove_profile(self,e, idx):
        """
        Handles the event for removing a profile.

        This method deletes the profile at the specified index and updates the UI accordingly.
        It also ensures that the selected profile is updated and the overlay is refreshed.

        Args:
            e (Event): The event triggered by the action to remove the profile.
            idx (int): The index of the profile to be removed.
        """
        self.config_class.delete_profile(self.config_class.index_to_name(idx))
        self.profile_ref.current.value = self.config_class.get_selected_profile_name()
        self.update_overlay()
        self.page.update()

    def update_overlay(self):
        """
        Updates the profile chooser overlay with the latest list of profiles.

        This method rebuilds the list of profile items when a profile's name is changed
        or a profile is deleted. It then updates the overlay content to reflect the
        changes and ensures the profile list is displayed correctly.
        """
        new_picker_items = self.create_list_items()
        new_content = ft.Stack([ft.Row([ft.Column(
            controls=[ft.Card(
                    ft.Container(
                        ft.ListView(
                            controls=new_picker_items,
                            height=self.calc_height(),
                            width=350,
                            spacing=10,
                            padding=10,
                        )
                    ,padding=15
                    )
                ,height=self.calc_height()+60
                )
            ],
            alignment=ft.MainAxisAlignment.CENTER,
        )],alignment=ft.MainAxisAlignment.CENTER),])
        self.profile_chooser_overlay.content = new_content

    def add_profile_pressed(self,e):
        """
        Handles the event for adding a new profile.

        Args:
            e (Event): The event triggered by the action to add a profile.
        """
        if self.gui.directory.is_lif:
            default = create_default_config()["Profiles"]["Lif"]
        else:
            default = create_default_config()["Profiles"]["Tif"]
        counter = 0
        new_name = "new Profile"

        while self.config_class.is_profile_existing(new_name):
            counter += 1
            new_name = "new Profile" + str(counter)

        self.config_class.add_profile(new_name,default["bf_channel"],default["mask_suffix"],default["channel_prefix"],default["diameter"])
        self.selected_profile_changed(e, self.config_class.name_to_index(new_name))
        self.update_overlay()
        self.page.update()

    def create_list_items(self):
        """
        Creates and returns a list of profile items with corresponding buttons for interaction.

        This method generates a list of items, where each item includes:
        -   A text field for displaying and editing the profile name.
        -   A button to select the profile by name.
        -   A delete button, which is only included if there are more than one profile.
        -   A edit button to activate the text field for profile renaming.

        Returns:
            List[ft.Row]: A list of rows containing the profile controls for interaction,
                          including buttons for deleting, editing, and selecting profiles.
        """
        self.name_items = self.create_name_items_profiles()
        if len(self.config_class.config["Profiles"]) > 1:
            return [
                ft.Row(
                    controls=[
                        self.name_items[i]["textfield"],
                        self.name_items[i]["button"],
                        ft.IconButton(
                            icon=ft.Icons.DELETE,
                            icon_color=ft.Colors.RED,
                            on_click=lambda e, i=i: self.remove_profile(e, i),
                        ),
                        ft.IconButton(
                            icon=ft.Icons.DRAW,
                            icon_color=ft.Colors.BLUE,
                            on_click=lambda e, i=i: self.text_field_activate(e, i),
                        ),
                    ],
                    alignment=ft.MainAxisAlignment.CENTER,
                )
                for i in reversed(range(len(self.config_class.config["Profiles"])))
            ]
        else:
            return [
                ft.Row(
                    controls=[
                        self.name_items[i]["textfield"],
                        self.name_items[i]["button"],
                        ft.IconButton(
                            icon=ft.Icons.DRAW,
                            icon_color=ft.Colors.BLUE,
                            on_click=lambda e, i=i: self.text_field_activate(e, i),
                        ),
                    ],
                    alignment=ft.MainAxisAlignment.CENTER,
                )
                for i in range(len(self.config_class.config["Profiles"]))
            ]


    def bf_updater(self, e):
        """
        Handles the event of updating the bright-field (BF) channel.

        This method validates the entered value for the BF channel. If the value is invalid,
        an error message is displayed to the user. A valid BF channel is expected to be not empty.

        Args:
            e (Event): The event triggered by the user interaction, containing the new BF channel value.
        """
        try:
            self.config_class.update_profile(self.config_class.get_selected_profile_name(),
                                             bf_channel=e.control.value)
            self.txt_bf_ref.current.color = None
            self.gui.directory.update_all_masks_check()
            self.gui.diameter_text.value = self.gui.average_diameter.get_avg_diameter()
            self.page.update()
            if self.gui.csp.image_id is not None:
                update_main_image(self.gui.csp.image_id, self.gui.csp.channel_id, self.gui, False)
            if not self.gui.csp.readout_running and not self.gui.csp.segmentation_running:
                self.gui.page.run_task(self.gui.directory.check_masks)
        except ValueError:
            self.page.open(ft.SnackBar(ft.Text("Bright field channel must be not empty!")))
            self.txt_bf_ref.current.color = ft.Colors.RED
            self.page.update()

    def ms_updater(self, e):
        """
        Handles the event of updating the mask suffix.

        This method checks if the entered mask suffix is valid. If the suffix is empty,
        an error message is displayed to the user. Otherwise, the mask suffix is updated
        for the selected profile.

        Args:
            e (Event): The event triggered by the user interaction, containing the new mask suffix value.
        """
        if e.control.value:
            self.config_class.update_profile(self.config_class.get_selected_profile_name(), mask_suffix=e.control.value)
            self.txt_ms_ref.current.color = None
            self.page.update()
        else:
            self.page.open(ft.SnackBar(ft.Text("Mask suffix must be not empty!")))
            self.txt_ms_ref.current.color = ft.Colors.RED
            self.page.update()

    def cp_updater(self,e):
        """
        Handles the event of updating the channel prefix.

        This method checks if the entered channel prefix is valid. If the prefix is empty,
        an error message is displayed to the user. Otherwise, the channel prefix is updated
        for the selected profile.

        Args:
            e (Event): The event triggered by the user interaction, containing the new channel prefix value.
        """
        if e.control.value:
            self.config_class.update_profile(self.config_class.get_selected_profile_name(), channel_prefix=e.control.value)
            self.txt_cp_ref.current.color = None
            self.page.update()
        else:
            self.page.open(ft.SnackBar(ft.Text("Channel prefix must be not empty!")))
            self.txt_cp_ref.current.color = ft.Colors.RED
            self.page.update()

    def d_updater(self,e):
        """
        Handles the event of updating the diameter.

        This method checks if the entered diameter is a valid decimal number. If the value is
        invalid or not greater than 0, an error message is displayed to the user. Otherwise,
        the diameter is updated for the selected profile.

        Args:
            e (Event): The event triggered by the user interaction, containing the new diameter value.
        """
        try:
            self.config_class.update_profile(self.config_class.get_selected_profile_name(), diameter=float(e.control.value))
            self.txt_d_ref.current.color = None
            self.page.update()
        except ValueError:
            self.page.open(ft.SnackBar(ft.Text("Diameter only allows decimals numbers, greater than 0!")))
            self.txt_d_ref.current.color = ft.Colors.RED
            self.page.update()

    def create_profile_container(self):
        """
        Creates and returns the profile container for the GUI.

        This method generates a container that includes:
        -   A row with the current selected profile and a button to open the profile chooser overlay.
        -   Text fields for modifying various attributes of the selected profile:
            -   Bright-Field Channel
            -   Mask Suffix
            -   Channel Prefix
            -   Diameter

        Each text field is associated with an updater method to handle value changes and validate inputs.

        Returns:
            ft.Container: The container holding the text fields for modifying selected profile attributes and the button to open the overlay.
        """
        #--------------------------------------
        #creates the TextFields for the diffrent attributes of a profile
        tf_bf = ft.TextField(
            label="Bright-Field Channel:",
            border_color=ft.Colors.BLUE_ACCENT,
            value=self.config_class.get_bf_channel(),
            ref=self.txt_bf_ref,
            on_blur=lambda e: self.bf_updater(e),
            width=200,
            height=60,
        )

        tf_ms = ft.TextField(
            label="Mask Suffix:",
            border_color=ft.Colors.BLUE_ACCENT,
            value=self.config_class.get_mask_suffix(),
            ref=self.txt_ms_ref,
            on_blur=lambda e: self.ms_updater(e),
            width=200,
            height=60,
        )

        tf_cp = ft.TextField(
            label="Channel Prefix:",
            border_color=ft.Colors.BLUE_ACCENT,
            value=self.config_class.get_channel_prefix(),
            ref= self.txt_cp_ref,
            on_blur=lambda e: self.cp_updater(e),
            width=200,
            height=60,
        )

        tf_d = ft.TextField(
            label="Diameter:",
            border_color=ft.Colors.BLUE_ACCENT,
            value=self.config_class.get_diameter(),
            ref=self.txt_d_ref,
            on_blur=lambda e: self.d_updater(e),
            width=200,
            height=60,
        )



        #creates the Row with the current selected Profile as button and when you click on it,
        #it opens the profile_chooser_overlay
        profiles_row = ft.Row(
            tight=True,
            controls=[
                ft.Text("Profile:", size=18),
                ft.TextButton(
                    content=ft.Text(self.config_class.get_selected_profile_name(), size=18, ref=self.profile_ref),
                    style=ft.ButtonStyle(color=ft.Colors.BLUE),
                    on_click=lambda e: self.profile_chooser_overlay.open()
                ),
                ft.IconButton(
                    icon=ft.Icons.LIBRARY_ADD_ROUNDED,
                    content=ft.Text("Add Profile", size=18),
                    tooltip="Add new profile",
                    on_click=lambda e: self.add_profile_pressed(e)
                ),

            ],
        )

        #creates the final Container that is displayed in the GUI
        return ft.Container(ft.Column([
                            profiles_row,
                            ft.Row([tf_bf, tf_cp]),
                            ft.Row([tf_ms, tf_d])]),padding=10)



import asyncio
from typing import Any

import flet as ft

class PageOverlay(ft.Stack):
    """
    Overlay which gets placed above the normal page with half transparent background
    and can be dismissed when clicking on the background when modal is False.

    Attributes:
        page: the page of Flet
        content: the content which gets viewed in front of the background
        on_dismiss: callback when the overlay gets dismissed
        modal: Whether this bottom sheet can be dismissed/closed by clicking the area outside of it.
    """
    def __init__(self,page: ft.Page,content: ft.Stack = None,on_dismiss = None,modal = False):
        super().__init__()
        self.page = page
        self.controls = []
        self._content: ft.Stack | None = None
        self.on_dismiss: Any | None = on_dismiss
        self.modal = modal
        self._background = self.create_background()
        self.content = content
        self.container = ft.Container(content=self,
                                      animate_opacity=ft.Animation(duration=300,
                                                                   curve=ft.AnimationCurve.LINEAR_TO_EASE_OUT),
                                      animate=ft.Animation(duration=300, curve=ft.AnimationCurve.LINEAR_TO_EASE_OUT), visible=False,opacity=0.0)
        page.overlay.append(self.container)

    @property
    def content(self) -> ft.Stack | None:
        return self._content

    @content.setter
    def content(self, new_content: ft.Stack):
        if new_content is not None:
            if self._content is not None:
                self.controls.remove(self._content)
            self.controls.append(new_content)
            self._content = new_content

    def open(self):
        self.page.run_task(self._open)

    def close(self):
        self.page.run_task(self._close)

    async def _open(self):
        self.container.visible = True
        self.page.update()
        await asyncio.sleep(0.14)
        self.container.opacity = 1.0
        self.container.update()

    async def _close(self):
        self.container.opacity = 0.0
        self.container.update()
        await asyncio.sleep(0.14)
        self.container.visible = False
        self.page.update()

    def create_background(self):
        def bg_click(e):
            if not self.modal:
                self.close()
                if self.on_dismiss is not None:
                    self.on_dismiss(e)

        background = ft.GestureDetector(
            mouse_cursor=ft.MouseCursor.BASIC,
            on_tap=bg_click,
            content=ft.Container(
            expand=True,
            bgcolor=ft.Colors.with_opacity(0.5, ft.Colors.BLACK),
            )
        )
        self.controls.append(background)
        return background
